# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (C) 2013 Rémi Pannequin, Centre de Recherche en Automatique de Nancy remi.pannequin@univ-lorraine.fr
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

"""
This module contains utilities to load and save emulica models in XML format.
Only module *configuration* is written : the state of a particular module 
is ignored.
"""
import os.path
from xml.etree.ElementTree import ElementTree, Element, SubElement
#ElementTree._namespace_map["http://www.w3.org/2001/XMLSchema"] = 'xs'
from emulica.core import emulation, properties
import logging, zipfile, pickle

logger = logging.getLogger('emulica.emuML')

class EmuFile(object):
    """This class enable to read or write Emulica files."""
    def __init__(self, filename, mode='r', parent_model=None, name='main'):
        """Create a new instance of a EmuFile file.

        Arguments:
            filename -- the name of the .emu file to open
            mode -- the open mode 'r' for reading, 'w' for writing
            parent_model -- the parent Emulation model, if current model is a 
                            submodel
            name -- the name of the model to load, if it is a submodel

        Raises:
            IOError if a file is absent or not readeable/writable
            FileNotFoundError if a submodel is absent
            EmuMLError if mode is not r or w, or if file in not a valid emu file.
        """
        self.filename = filename
        self.name = name
        self.mode = mode
        if mode == 'r' or mode == 'w':
            self.zfile = zipfile.ZipFile(filename, mode)
            if mode == 'r':
                namelist = self.zfile.namelist()
                if 'emulation.xml' in namelist:
                    model_xml = self.zfile.read('emulation.xml')
                    self.efile = EmulationParser(model_xml,
                                                 name = name,
                                                 path = self.filename,
                                                 parent = parent_model)
                else:
                    raise EmuMLError(_("unable to find emulation.xml"))
                if 'control.py' in namelist:
                    self.control = self.zfile.read('control.py')
                else:
                    raise EmuMLError(_("unable to find control.py"))
                if 'props.db' in namelist:
                    self.__props = pickle.loads(self.zfile.read('props.db'))
                else:
                    raise EmuMLError(_("unable to find props.db"))
        else:
            raise EmuMLError(_("Invalid opening mode, only 'r' (for reading) and 'w' (for writing) are valid modes."))

    def read(self, parent=None, name='main'):
        """Read the content of a model. This class parse the emulation.xml file
        contained in the emu file, and also parse and loads submodels. NB:
        model and submodels properties can be retrieved using get_properties()
        
        Returns:
            a tuple containing the emulation model and the control file
        Raises:
            EmuMLError -- if file was not open in read mode or if the emu
                           file is not correct
            IOError -- if a submodel could not be read
        """
        if self.mode != 'r':
            raise EmuMLError(_("File opened in write mode, cannot read."))
        self.efile.parse()
        return (self.efile.model, self.control)

    def get_properties(self):
        """Return a dictionary of the model and submodels properties. Models are
        identified by their names (the top-model name is 'main')"""
        props = dict()
        props[self.name] = self.__props
        for (name, gsf) in self.efile.submodels.items(): 
            for (name, prop) in gsf.get_properties().items():
                props[name] = prop
        return props

    def write(self, model, control, properties):
        """Write the emu file element into a emu file.

        Arguments:
            model -- the emulation model
            control -- the control system
            properties -- the model properties

        """
        if self.mode != 'w':
            raise EmuMLError(_("File opened in read mode, cannot write."))
        self.write_model(model)
        self.write_control(control)
        self.write_properties(properties)

    def write_model(self, model):
        """Write the emmulation model in the emu file.

        Arguments:
            model -- the model to write

        """
        efile = EmulationWriter(model)
        self.zfile.writestr('emulation.xml', efile.write())

    def write_control(self, control):
        """Write the control string in the emu file.

        Arguments:
            control -- the code (as text) of the control system

        """
        self.zfile.writestr('control.py', control)

    def write_properties(self, props):
        """Write the props dictionary in the emu file (using pickle).

        Arguments:
            props -- a dictionary of model properties

        """
        props_str = pickle.dumps(props) #get emulica config (using pickle)
        self.zfile.writestr('props.db', props_str)

    def close(self):
        """Close the emu file. Any operation is invalid after."""
        try:
            for zinfo in self.zfile.infolist():
                zinfo.external_attr = 600 << 16
        finally:
            self.zfile.close()


class EmulationWriter(object):
    """This class write an emulation model as an xml string"""
    def __init__(self, model):
        """Create a new instance of a emuML.EmulationFile.

        Arguments:
            model -- the model from which extract modules

        """
        self.model = model

    def write(self):
        """Return an XML string that correspond to model. NB: submodels are not
        outputed."""
        from xml.etree.ElementTree import tostring
        root = Element("emulationModel")
        input_root = SubElement(root, "interface")
        for (name, (module_name, property_name)) in self.model.inputs.items():
            input_elt = SubElement(input_root, "input")
            input_elt.attrib['name'] = name
            input_elt.attrib['module'] = module_name
            input_elt.attrib['property'] = property_name
        mod_root = SubElement(root, "modules")
        #model structure
        for (name, mod) in self.model.modules.items():
            if mod.is_model():
                mod_root.append(self.marshall_submodel(mod))
            else:
                mod_root.append(self.marshall_module(mod))
        tree = ElementTree(root)
        return tostring(root)

    def marshall_submodel(self, module):
        """Return a XML element that represent a submodel inclusion (and its properties)"""
        #create a 
        submodel_elt = Element("submodel")
        submodel_elt.attrib['name'] = module.name
        submodel_elt.attrib['path'] = module.path#TODO: modify mod.path to a relative path ?
        for (name, prop) in module.properties.items():
            prop_elt = SubElement(submodel_elt, "property")
            prop_elt.attrib["name"] = name
            self.marshall_value(prop_elt, prop)
            logger.debug("marshaling attribute {0} with value {1} for module {2}".format(name, prop, module.name))
        return submodel_elt

    def marshall_module(self, module):
        """Return a XML element that represents the module.

        Arguments:
            module -- the module to marshall

        """
        mod_elt = Element("module")
        mod_elt.attrib["name"] = module.name
        mod_elt.attrib["type"] = module.__class__.__name__
        for (name, prop) in module.properties.items():
            prop_elt = SubElement(mod_elt, "property")
            prop_elt.attrib["name"] = name
            self.marshall_value(prop_elt, prop)
            logger.debug("marshaling attribute {0} with value {1} for module {2}".format(name, prop, module.name))
        return mod_elt


    def marshall_value(self, root, value):
        """Append one or several elements to root. If value is None, nothing is
        appended, if value is a list, an element is append for each object of
        the list, else one element is appended.

        Arguments:
            root -- the element to which append the created element
            value -- the property to marshall

        """
        if value is None:
            logger.warning("marshalling a null value")
        elif type(value) == list:
            list_root = SubElement(root, 'value-list')
            for v in value:
                self.marshall_value(list_root, v)
        else:
            if value.__class__.__name__ == 'ProgramTable':
                #program-table element
                self.marshall_prog(root, value)
            elif value.__class__.__name__ == 'SetupMatrix':
                #setup-table element
                self.marshall_setup(root, value)
            elif value.__class__.__name__ == 'ChangeTable':
                #change-table element
                self.marshall_change(root, value)
            elif 'fullname' in dir(value):
                #reference element
                elt = SubElement(root, 'reference')
                elt.text = value.fullname()
            else:
                #value element
                elt = SubElement(root, 'value')
                elt.text = str(value)
                #TODO: manage type ?
                #elt.attrib["type"] = str(type(value))

    def marshall_prog(self, root, table):
        """Append an Element that represents a program table to element 
        prog_root.

        Arguments:
            root -- the program table to marshall
            table -- the element to which append the created element

        """
        prog_root = SubElement(root, 'program-table')
        #get program structure, write as attribute
        schema = [name for (name, display) in table.program_keyword]
        schema_str = '['+','.join(schema)+']'
        prog_root.attrib["schema"] = schema_str
        for (name, p) in table.items():
            prog = SubElement(prog_root, "program")
            prog.attrib["name"] = name
            prog.attrib["delay"] = str(p.time_law)
            if p.transform and 'items' in dir(p.transform):
                for (n, e) in p.transform.items():
                    elt = SubElement(prog, 'transform')
                    elt.attrib['name'] = n
                    self.marshall_value(elt, e)

    def marshall_setup(self, root, table):
        """Append a setup-table Element that represents a setup table to element 
        root.

        Arguments:
            root -- the element to which append the created element
            table -- the setup table to marshall

        """
        setup_root = SubElement(root, 'setup-table')
        setup_root.attrib["default_delay"] = str(table.default_time)
        for (initial, final, delay) in table.items():
            setup = SubElement(setup_root, 'setup')
            setup.attrib['initial'] = initial
            setup.attrib['final']  = final
            setup.attrib['delay'] = str(delay)


    def marshall_change(self, root, table):
        """Append a change-table Element that represents a change table to element 
        root.

        Arguments:
            root -- the element to which append the created element
            table -- the setup table to marshall
           """
        change_root = SubElement(root, 'change-table')
        for (name, value) in table.items():
            change = SubElement(change_root, 'change')
            change.attrib["property"] = name
            self.marshall_value(change, value)


class EmulationParser(object):
    """This class can be used to get an Emulation Model from an xml file."""
    def __init__(self, string, model = None, parent = None, name = 'main', path = None):
        """Create a new instance of a emuML.EmulationParser. When the object is 
        created, the string is parsed, and submodels list is built. Beware: the
        submodels files are opened at this stage ! If the model argument is 
        specified, it is used to load modules into, else a new model is created 
        using the kwargs.
        
        Arguments:
            string -- the model into which load modules or from which extract modules
            model -- the model to load modules into
            *kwargs -- keywords argumenst to be passed to the Model constructor
        Raises:
            ???Error -- if string is not well formed   
         
        """
        from xml.etree.ElementTree import fromstring
        self.tree = fromstring(string)
        self.model = model or emulation.Model(model = parent, name = name, path = path)
        self.submodels = dict()
        mod_root = self.tree.find("modules")
        for submodel_elt in mod_root.findall("submodel"):
            #try loading emu file for every submodels
            sub_path = submodel_elt.get('path')
            #TODO If sub_path is a relative pathname, interpret it where the current model is
            if not os.path.isabs(sub_path):
                base = os.path.dirname(path)
                sub_path = os.path.join(base, sub_path)
            sub_name = submodel_elt.get('name')
            gsf = EmuFile(sub_path, 'r', parent_model = self.model, name = sub_name)
            #TODO: if opening fails, add name to a list of broken submodels
            self.submodels[sub_name] = gsf
        self.renaming = dict()
    
    def load_submodels(self):
        """Load submodels in the model"""
        for (name, gsf) in self.submodels.items():
            (submodel, subcontrol) = gsf.read()
            #compile and register control in model
            compile_control(submodel, subcontrol)    
    
    def parse(self):
        """Load a set of modules from an XML string or treeset into the model, 
        and Return the list of created modules.
        
        Arguments:
            string -- an XML string that represent the model.
            
        Return 
            the list of created modules
        
        """
        
        mod_root = self.tree.find("modules")
        mod_list = []
        properties_elt = dict()
        #loads submodels first (if any)
        self.load_submodels()
        #loads modules
        for mod_elt in mod_root.findall("module"):
            mod_type = mod_elt.get("type")
            name = mod_elt.get("name")
            if name in self.model.modules:
                new_name = mod_type+str(len(self.model.modules))
                self.renaming[name] = new_name
                logger.warning("renaming module: " + name + " to: "+new_name)
                name = new_name
            logger.info("creating module: " + name + " of type: "+mod_type)
            mod_class = getattr(emulation, mod_type)
            mod = mod_class(self.model, name)
            mod_list.append(mod)
            properties_elt[mod] = mod_elt.findall('property')
            
        #build model structure
        for mod, prop_list in properties_elt.items():
            for prop_elt in prop_list:
                name = prop_elt.get('name')
                value = self.parse_child_as_value(mod.properties, name, prop_elt)
                #print "property {0} = {1}".format(name, value)
                mod.properties[name] = value
        #get inputs
        interface_root = self.tree.find('interface')
        if not interface_root is None:
            for input_elt in interface_root.findall('input'):
                name = input_elt.get('name')
                module = input_elt.get('module')
                prop = input_elt.get('property')
                self.model.inputs[name] = (module, prop)
            #if model has parent, call apply_inputs
            if not self.model.is_main:
                self.model.apply_inputs()
        #load submodel properties
        for submodel_elt in mod_root.findall('submodel'):
            submodel = self.model.get_module(submodel_elt.get('name'))
            for prop_elt in submodel_elt.findall('property'):
                name = prop_elt.get('name')
                value = self.parse_child_as_value(submodel.properties, name, prop_elt)
                #print "property {0} = {1}".format(name, value)
                submodel.properties[name] = value
        return mod_list

    def parse_child_as_value(self, props, root_prop_name, element):
        """Parse the children of element as values (or list of values)"""
        children = element.getchildren()
        if len(children) == 0:
            logger.warning("returning None when parsing element")
            return None
        elif len(children) == 1:
            return self.parse_value(props, root_prop_name, children[0])
        else:
            logger.warning("wrong number of values")

    def parse_value(self, props, root_prop_name, element):
        """Parse a module property and return the resulting object
        
        Arguments:
            module -- the module which the property belongs
            element -- the Element to parse
        
        Returns:
            the parsed value
            
        """
        if element.tag == 'value':
            elt_type = element.get("type")
            #TODO: get type and cast to type
            try:
                return eval(element.text)
            except (NameError, AttributeError):
                return element.text
                
        elif element.tag == 'value-list':
            children = element.getchildren()
            value = list()
            for child in children:
                value.append(self.parse_value(props, root_prop_name, child))
            return value
        elif element.tag == 'reference':
            name = element.text
            if name in self.renaming:
                name = self.renaming[name]
            return self.model.get_module(name)
        elif element.tag == 'program-table':
            return self.parse_prog(props, root_prop_name, element)
        elif element.tag == 'setup-table':
            return self.parse_setup(props, root_prop_name, element)
        elif element.tag == 'change-table':
            return self.parse_change(props, root_prop_name, element)
        else:
            #error case
            logger.warning("errror: unknow tag {0}".format(element.tag))

    def parse_prog(self, props, root_prop_name, element):
        """Parse a program table and return the resulting object.

        Arguments:
            element -- the Element to parse

        Returns:
            the parsed program table (type dictionary)

        """
        #get the program table schema
        schema_str = element.get("schema")
        context = {'source': ('source', properties.Display(properties.Display.REFERENCE, _("Source"))),
                   'destination': ('destination',properties.Display( properties.Display.REFERENCE, _("Destination"))),
                   'change': ('change', properties.Display(properties.Display.PHYSICAL_PROPERTIES_LIST, _("Physical changes")))}
        if schema_str is None:
            if 'program_keyword' in dir(props.owner):
                schema = props.owner.program_keyword
            else:
                logger.warning(_("unable to determine program schema"))
                schema = []
        else:
            schema = eval(schema_str, globals(), context)
        table = properties.ProgramTable(props, root_prop_name, schema)
        for program in element.findall('program'):
            delay = program.get("delay")
            #get transform
            transform = dict()
            for elt in program.findall('transform'):
                value = self.parse_child_as_value(props, root_prop_name, elt)
                transform[elt.get("name")] = value      
            name = program.get("name")
            assert(not name is None)
            table.add_program(name, delay, transform)
        return table

    def parse_setup(self, props, root_prop_name, element):
        """Parse a setup table and return the resulting object.

        Arguments:
            element -- the Element to parse

        Returns:
            the parsed SetupTable

        """
        default_delay = element.get("default_delay")
        setup = properties.SetupMatrix(props, default_delay, root_prop_name)
        for s in element.findall('setup'):
            delay = eval(s.get("delay"))
            init = s.get("initial")
            final = s.get("final")
            assert(not init is None)
            assert(not final is None)
            setup.add(init, final, delay)
        return setup

    def parse_change(self, props, root_prop_name, element):
        """Parse a change table and return the resulting object"""
        table = properties.ChangeTable(props, root_prop_name)
        for child in element.findall('change'):
            name = child.get("property")
            assert(not name is None)
            value = self.parse_child_as_value(props, root_prop_name, child)
            table[name] = value
        return table

def compile_control(model, source, **kwargs):
    """compile code in the control buffer. The global environment that is available
    in the control code is the package emulation, and classes Request and Report.
    The control code must provide a function initialize_control that registers the
    control modules in the model.
    
    Arguments:
        model -- the model into which load the control
        source -- the source code of the control system

    Any other keyword argument can be added : these kw arguments will be passed
    to the initialize_control method of the source. 

    Returns: 
        True if no error are found.

    Raises:
        SyntaxError -- if there is syntax error in the source
        EmuMLError -- if init function has not been declared

    """
    code = compile(source,'<control buffer>', 'exec')
    g = globals()
    env = dict()
    env['emulation'] = emulation
    env['Request'] = emulation.Request
    env['Report'] = emulation.Report
    l = dict()
    exec(code, env, l)
    logger.info(_("control code compilation sucessfull"))
    if not 'initialize_control' in l:
        raise EmuMLError(_("initialization function have not been implemented"))
    model.control_classes = []
    #This function come from the code we just compiled and executed
    init_function = l['initialize_control']
    init_function(l, model, **kwargs)

def save_modules(model, modules):
    """Return an XML string that represent these modules. Used to cut or copy 
    modules.
    
    Arguments:
        model -- the model (context, useful for references)
        modules -- the list of module to marshall
        
    """
    from xml.etree.ElementTree import tostring
    writer = EmulationWriter(model)
    top = Element("emulationClip")
    mod_root = SubElement(top, "modules")
    #model structure
    for mod in modules:
        mod_root.append(writer.marshall_module(mod))
    return tostring(top)

def load_modules(model, elt):
    """Convenience function to load a set of modules from an XML string in the 
    model. Used to paste modules.
    
    Arguments:
        model -- the model where the module are loaded
        elt -- the XML string that describes the modules
    """
    parser = EmulationParser(elt, model = model)
    parser.parse()

def save(model, filename=None):
    """Convenience function to save a model to a file. If filename is not set, 
    the resulting xml string is returned."""
    writer = EmulationWriter(model)
    if filename is None:
        return writer.write()
    else:
        try:
            efile = open(filename, 'w')
            content = writer.write()
            efile.write(content)
        finally:
            efile.close()

def load(filename):
    """Convenience function that returns a new Model from a file"""
    f = open(filename, 'r')
    content = f.read()
    f.close()
    efile = EmulationParser(content, path = filename)
    efile.parse()
    return efile.model

def parse_request(message):
    """Parse a XML message and return a Request instance
    
    Returns:
        a emulation.Request instance
        
    Raises:
        EmuMLError -- if the request doesn't respects schema
        
    """
    from xml.etree.ElementTree import fromstring
    try:
        root = fromstring(message)
    except Exception as message:
        raise EmuMLError(message)
    attributes = dict()
    elt = root.find('who')
    if elt == None: 
        raise EmuMLError("request message does not specify who must execute it")
    who = elt.text
    elt = root.find('what')
    if elt == None: 
        raise EmuMLError("request message does not specify what to execute")
    what = elt.text
    rq = emulation.Request(who, what)
    elt = root.find("how")
    if not elt == None:
        for e in elt.findall("element"):
            try:
                value = e.text
            except (SyntaxError, NameError):
                value = e.text
            rq.how[e.attrib["name"]] = value
    elt = root.find("where")
    if not elt == None:
        rq.where = elt.text
    elt = root.find("when")
    if not elt == None:
        try:
            rq.when = float(elt.text)
        except ValueError:
            raise EmuMLError("""'{0}' could not be recognized as a date (float)""".format(elt.text))
    elt = root.find("why")
    if not elt == None:
        rq.why = elt.text
    return rq

def parse_report(message):
    """Parse a XML message and return a Report instance
    
    Returns:
        a emulation.Report instance
        
    Raises:
        EmuMLError -- if the request doesn't respects schema
        
    """
    from xml.etree.ElementTree import fromstring
    try:
        root = fromstring(message)
    except Exception as message:
        raise EmuMLError(message)
    attributes = dict()
    elt = root.find('who')
    if elt == None: 
        raise EmuMLError("request message does not specify who must execute it")
    who = elt.text
    elt = root.find('what')
    if elt == None: 
        raise EmuMLError("request message does not specify what to execute")
    what = elt.text
    rp = emulation.Report(who, what)
    elt = root.find("how")
    if not elt == None:
        for e in elt.findall("element"):
            try:
                value = eval(e.text)
            except (SyntaxError, NameError):
                value = e.text
            rp.how[e.attrib["name"]] = value
    elt = root.find("where")
    if not elt == None:
        rp.where = elt.text
    elt = root.find("when")
    if not elt == None:
        try:
            rp.when = float(elt.text)
        except ValueError:
            raise EmuMLError("""'{0}' could not be recognized as a date (float)""".format(elt.text))
    elt = root.find("why")
    if not elt == None:
        rp.why = elt.text
    return rp

def write_report(report):
    """Return an XML message from a report object"""
    from xml.etree.ElementTree import tostring
    root = Element("report")
    for attr in ['who', 'what', 'where', 'why', 'when']:
        value = getattr(report, attr)
        if value:
            elt = SubElement(root, attr)
            elt.text = str(value)
    how_elt = SubElement(root, 'how')
    for (name, value) in report.how.items():
        elt = SubElement(how_elt, 'element')
        elt.attrib['name'] = name
        elt.text = str(value)
    return tostring(root)

def write_request(request):
    """Return an XML message from a request object"""
    from xml.etree.ElementTree import tostring
    root = Element("request")
    for attr in ['who', 'what', 'when', 'where', 'why']:
        value = getattr(request, attr)
        if value:
            elt = SubElement(root, attr)
            elt.text = str(value)
    how_elt = SubElement(root, 'how')
    for (name, value) in request.how.items():
        elt = SubElement(how_elt, 'element')
        elt.attrib['name'] = name
        elt.text = str(value)
    return tostring(root)


class EmuMLError(Exception):
    """This class is used to deal with exception ancountrered when reading or 
    writing a file."""



#!/usr/bin/python
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (C) 2013 Rémi Pannequin, Centre de Recherche en Automatique de Nancy remi.pannequin@univ-lorraine.fr
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

"""Submodel testing: use a model input in the control system"""

import unittest

import logging
from emulica.core import set_up_logging
set_up_logging(logging.ERROR)

import util
util.set_path()

from emulica.core.emulation import *
from emulica.core.properties import SetupMatrix

EXP_RESULT_PRODUCT = [(1, [(3, 7, 'cell.machine', 'p1')], [(0, 'cell.source'),
                                                           (0, 'cell.transporter'),
                                                           (2, 'cell.espaceMachine'),
                                                           (7, 'cell.transporter'),
                                                           (9, 'cell.sink')], 0, 9),
                      (2, [(21, 27, 'cell.machine', 'p3')], [(0, 'cell.source'),
                                                             (17, 'cell.transporter'),
                                                             (19, 'cell.espaceMachine'),
                                                             (27, 'cell.transporter'),
                                                             (29, 'cell.sink')], 0, 29),
                      (3, [(40, 45, 'cell.machine', 'p2')], [(0, 'cell.source'),
                                                             (37, 'cell.transporter'),
                                                             (39, 'cell.espaceMachine'),
                                                             (45, 'cell.transporter'),
                                                             (47, 'cell.sink')], 0, 47),
                      (4, [], [(0, 'cell.source')], 0, 50)]
                      
EXP_RESULT_RESOURCE = [[(0, 0, 'setup'),
                        (0, 2, 'load'),
                        (7, 7, 'setup'),
                        (7, 9, 'unload'),
                        (17, 17, 'setup'),
                        (17, 19, 'load'),
                        (27, 27, 'setup'),
                        (27, 29, 'unload'),
                        (37, 37, 'setup'),
                        (37, 39, 'load'),
                        (45, 45, 'setup'),
                        (45, 47, 'unload')],
                       [(2, 3, 'setup'),
                        (3, 7, 'p1'),
                        (19, 21, 'setup'),
                        (21, 27, 'p3'),
                        (39, 40, 'setup'),
                        (40, 45, 'p2')]]


EMULATE_UNTIL = 50;

class ControlCreate:
    def run(self, model):
        n = 0
        i = 0
        pType = ['type1', 'type2', 'type3']
        createModule = model.get_module("create")
        while n < 4:
            m = Request("create", "create",params={'productType':pType[i]})
            yield createModule.request_socket.put(m)
            i = (i+1)%3
            n += 1

class ControlDispose:
    def run(self, model):
        dispose = model.get_module('dispose')
        obs = model.get_module('obsSink')
        rp_obs = obs.create_report_socket()
        while True:
            yield rp_obs.get()
            yield dispose.request_socket.put(Request("dispose", "dispose"))

class ControlCell:
    def run(self, model):
        prog = {'type1':'p1','type2':'p3','type3':'p2'}
        sp = model.get_module("transporter")
        machine = model.get_module("machine")
        rp_machine = machine.create_report_socket()
        obs1 = model.get_module("obsSource")
        rp_obs1 = obs1.create_report_socket()
        obs2 = model.get_module("obsMachine")
        rp_obs2 = obs2.create_report_socket()
        while True:
            ##attente de l'arrivée d'un pièce
            ev = yield rp_obs1.get()
            rq = Request("transporter","move",params={'program':'load'})
            yield sp.request_socket.put(rq)
            ##pièce prête
            ev = yield rp_obs2.get()
            p = prog[ev.how['productType']]
            yield machine.request_socket.put(Request("machine","setup", params={"program":p}))
            ##début process
            yield machine.request_socket.put(Request("machine","make"))
            ##attente fin process
            fin = False
            while not fin:
                ev = yield rp_machine.get()
                fin = ev.what=="idle"
            ##déchargement
            yield sp.request_socket.put(Request("transporter", "move", params={"program":'unload'}))
            #print "waiting %s" % model['delay']
            yield model.get_sim().timeout(model['delay'])
             

def create_submodel(parent, name, delay):
    model = Model(model = parent, name = name, path = 'cell.gseme')
    source = Holder(model, "source")
    sink = Holder(model, "sink")
    espaceMachine = Holder(model, "espaceMachine")
    PushObserver(model, "obsSource", "source-ready", observe_type = False, holder = source)
    PushObserver(model, "obsMachine", "machine-ready", holder = espaceMachine)
    sp = SpaceAct(model, "transporter")
    sp.add_program('load', 2, {'source':source, 'destination':espaceMachine})
    sp.add_program('unload', 2, {'source':espaceMachine, 'destination':sink})
    machine = ShapeAct(model, "machine", espaceMachine)
    machine.add_program('p1', 4)
    machine.add_program('p2', 5)
    machine.add_program('p3', 6)
    m = SetupMatrix(machine.properties, 1)
    m.add('p1','p3',2)
    m.add('p3','p1',3)
    machine['setup'] = m
    model.register_control(ControlCell)
    model.properties['delay'] = delay


def get_model():
    model = Model()
    create_submodel(model, "cell", 10)
    source = model.get_module("cell.source")
    CreateAct(model, "create", source)
    sink = model.get_module("cell.sink")
    DisposeAct(model, "dispose", sink)
    PushObserver(model, "obsSink", "sink-ready", holder = sink)
    model.register_control(ControlCreate)
    model.register_control(ControlDispose)
    return model



class TestSim16(unittest.TestCase):
    def setUp(self):
        print(self.id())

    def test_ModelCreate(self):
        get_model()

    def test_Start(self):
        model = get_model()
        model.emulate(until = EMULATE_UNTIL)

    def test_RunResults(self):
        model = get_model()
        model.emulate(until = EMULATE_UNTIL)
        result_product = [(pid, 
                       p.shape_history, 
                       p.space_history,
                       p.create_time, 
                       p.dispose_time) for (pid, p) in model.products.items()]
        result_resource = [model.get_module("cell.transporter").trace, model.get_module("cell.machine").trace]
        self.assertEqual(result_product, EXP_RESULT_PRODUCT)
        self.assertEqual(result_resource, EXP_RESULT_RESOURCE)


if __name__ == '__main__':    
    unittest.main()




#!/usr/bin/python
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (C) 2013 Rémi Pannequin, Centre de Recherche en Automatique de Nancy remi.pannequin@univ-lorraine.fr
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

"""
Create -> holder -> space -> holder
"""

import unittest

import logging
from emulica.core import set_up_logging
set_up_logging(logging.ERROR)

import util
util.set_path()

from emulica.core.emulation import *

EXP_RESULT_PRODUCT = [(1, [], [(0, 'h1'), (0, 'space1'), (2, 'h2')], 0, 41),
                      (2, [], [(10, 'h1'), (10, 'space1'), (12, 'h2')], 10, 41),
                      (3, [], [(20, 'h1'), (20, 'space1'), (22, 'h2')], 20, 41),
                      (4, [], [(30, 'h1'), (30, 'space1'), (32, 'h2')], 30, 41),
                      (5, [], [(40, 'h1'), (40, 'space1')], 40, 41)]                  
EXP_RESULT_RESOURCE = [(0, 0, 'setup'), 
                       (0, 2, 'p1'), 
                       (10, 12, 'p1'), 
                       (20, 22, 'p1'), 
                       (30, 32, 'p1'), 
                       (40, 41, 'p1')]
                       
EMULATE_UNTIL = 41;

class ControlCreate:
    def run(self, model):
        n = 0
        createModule = model.modules["create1"]
        report = createModule.create_report_socket()
        while n < 10:
            m = Request("create1", "create")
            yield createModule.request_socket.put(m)
            yield report.get()
            yield model.get_sim().timeout(10)


class ControlSpace:
    def run(self, model):
        sp = model.modules["space1"]
        obs1 = model.modules["observer1"]
        report = obs1.create_report_socket()
        while True:
            ev = yield report.get()
            #print ev
            rq = Request("space1","move",params={'program':'p1'})
            yield sp.request_socket.put(rq)
 
    
class MonitorSpace:
    def run(self, model):
        sp = model.modules["space1"]
        report = sp.create_report_socket()
        while True:
            yield report.get()
            

def get_model():
    model = Model()
    h1 = Holder(model, "h1")
    h2 = Holder(model, "h2")
    obs1 = PushObserver(model, "observer1", "ev1", holder = h1)
    c = CreateAct(model, "create1", h1)
    sp = SpaceAct(model, "space1")
    sp.add_program('p1', 2, {'source':h1, 'destination':h2})
    initialize_control(model)
    return model

def initialize_control(model):
    model.register_control(ControlCreate)
    model.register_control(ControlSpace)
    model.register_control(MonitorSpace)
  

class TestSim2(unittest.TestCase):
    def setUp(self):
        print(self.id())
    
    def test_ModelCreate(self):
        get_model()

    def test_Start(self):
        model = get_model()
        model.emulate(until = EMULATE_UNTIL)

    def test_RunResults(self):
        model = get_model()
        model.emulate(until = EMULATE_UNTIL)
        result_product = [(pid, p.shape_history, 
                       p.space_history, 
                       p.create_time, 
                       p.dispose_time) for (pid, p) in model.products.items()]
        result_resource = model.modules["space1"].trace
        self.assertEqual(result_product, EXP_RESULT_PRODUCT)
        self.assertEqual(result_resource, EXP_RESULT_RESOURCE)


if __name__ == '__main__':    
    unittest.main()



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (C) 2013 Rémi Pannequin, Centre de Recherche en Automatique de
# Nancy remi.pannequin@univ-lorraine.fr
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE
"""
This module contain classes related to the creation, management and displaying of
properties in emulica.
"""

import random
import logging

from emulica.core import emulation

logger = logging.getLogger('emulica.properties')

class Registry(dict):
    """
    Properties can be attached to every products and module (including models).
    They basically consist in a (name, value) pair. Moreover, a PropertyDisplay
    can also be used. If so the property is said to be "displayable";
    such properties can be presented in the user interface, and modified.
    Properties that don't have any PropertyDisplay object associated are not
    displayed by the PropertyWindow UI.

    Properties can be a direct value (a str, bool, int or float) a module (ie a
    pointer to a module), a string that can be evaluated (in a pythonic syntax),
    or complex object that contain other (sub)properties (setup or program
    tables).

    A property can be accessed from its owner using square braquets [], using
    its name.
    For example, create['destination'] returns the value of the 'destination'
    property of module 'create'. Shape1['program_table']['program1'] is the
    property 'program1' of the program_table object (which is in turn the
    property 'program_table' of module 'Shape1')... Properties can be get/set
    using the square braquet syntax, in the same way a python dictionary can be
    used.

    module['prop'] return the non evaluated value of the property. To evaluate
    the property string, the evaluate_property function must be used, or
    alternativelly, the property must be called. Therefore if module['prop'] is
    "product['mass'] + 5", module['prop'](context) is 12, in context wherethe
    current product has a property that evaluates to 7.

    Classes:
        Registry -- a dict of properties, with dsiplay information
        SetupMatrix -- Models setup (transtion between programs), and associated time
        Program -- Models actions done by actuators
        Display -- display information for properties
    """

    def __init__(self, owner, rng):
        """Create a new Registry, where properties can be stored.

        Arguments:
            owner -- the module that own this property registry
            parent -- the parent registry (used to get evaluation context and to
            notify owner of property modification)
        """
        dict.__init__(self)
        self.owner = owner
        self.rng = rng
        self.displays = dict()
        self.__ordered_display = list()
        self.auto_eval = set()

    def __getitem__(self, name):
        """Return the property (unevaluated), or the evaluated value
        if the property has been marqued.
        """
        if name in self.auto_eval:
            return self.evaluate(name)
        return dict.__getitem__(self, name)

    def __setitem__(self, name, value):
        """call the method of the underling dict, and trigger the modules
        property changed signal."""
        dict.__setitem__(self, name, value)
        self.notify_owner(name)

    def notify_owner(self, prop_name):
        """Notify owner or parent of property change"""
        if 'emit' in dir(self.owner):
            self.owner.emit(emulation.Module.PROPERTIES_CHANGE_SIGNAL, prop_name, self.owner)

    def set_auto_eval(self, name, auto_eval=True):
        """Mark the property 'name' to be automatically evaluated"""
        if auto_eval:
            self.auto_eval.add(name)
        else:
            self.auto_eval.remove(name)

    def add_with_display(self, name, display_type, value=None, display_name=None):
        """Add a new property and set its display. If value is not specified or None,
        it is instancied from the default_value dict in class Display. If display_name
        is not specified or None, it take avlue name.

        Arguments:
            name -- the property name
            display_type -- the display type (int, see properties.Display)

        Keyword arguments:
            value -- the property's value
            display_name -- the name to use in dialogs
        """
        if value is None:
            if 'program_keyword' in dir(self.owner):
                schema = self.owner.program_keyword
            else:
                schema = list()
            value = Display.default_value[display_type](self.owner.properties, name, schema)
        display_name = display_name or name
        self[name] = value
        display = Display(display_type, display_name)
        self.__ordered_display.append(name)
        self.displays[name] = display

    def set_display(self, name, display):
        """Set the display of the property designed by 'name'.

        Arguments:
            name -- the name of the property
            display -- the properties.Display to set
        """
        self.displays[name] = display
        if not display in self.__ordered_display:
            self.__ordered_display.append(name)

    def get_display(self, name):
        """Get the display of the property designed by name.

        Arguments:
            name -- the name of the property

        Returns:
            the associated Display object, or None if the property is not
            displayable
        """
        return self.displays[name]

    def displayables(self):
        """return a list of (name, value, display) tuple of all the properties
        that have a display set."""
        result = list()
        for name in self.__ordered_display:
            result.append((name, self[name], self.displays[name]))
        return result

    def get(self, name):
        """Return the value of the prop without evaluation"""
        return dict.__getitem__(self, name)

    def evaluate(self, name, product=None):
        """Resolve reference to other properties, and return the value.
        """
        return self.eval_expression(self.get(name), product)

    def eval_and_set(self, name, value, product=None):
        """Evaluate value in the context of product, and set the result as prop
        name.
        Exemple product.properties.eval_and_set('mass', 'self['mass'] / 2')
        """
        self[name] = self.eval_expression(value, product)

    def eval_expression(self, expr, product=None):
        """Evaluate expression expr"""
        if type(expr) == str:
            context = dict()
            context['rng'] = self.rng
            if self.owner and 'model' in dir(self.owner):
                context['model'] = self.owner.model
            for (name, value) in self.items():
                context[name] = value
            if not product is None:
                context['product'] = product
            result = eval(expr, globals(), context)
        else:
            result = expr
        return result


class SetupMatrix(object):
    """A Setup Matrix record setup times between programs.

    When requesting a setup time for a (initial -> final) transition
        * if it exists, that precise transition is used
        * else, if there is a default for that final, it is used
        * else, the default setup time is used

    Attributes:
        default_time -- default setup time when no setup data have been found
    """
    def __init__(self, prop_registry, default_time=0, parent_prop_name='setup'):
        """Create a new instance of SetupMatrix

        Arguments:
            default_time -- default setup time (default = 0)
        """
        self.registry = prop_registry
        assert type(parent_prop_name) == str
        assert 'notify_owner' in dir(self.registry)
        self.parent_prop_name = parent_prop_name
        self.default_time = default_time
        self.__dest_default = dict()
        self.__dest_prog = dict()

    def add(self, initial_prog, final_prog, setup_time):
        """Add a new element in the matrix.

        Arguments:
            initial_prog -- the program at the begining of the setup
            final_prog -- the program at the end of the setup
            setup_time -- the setup delay
        """
        if final_prog not in  self.__dest_prog:
            self.__dest_prog[final_prog] = dict()
        self.__dest_prog[final_prog][initial_prog] = setup_time
        self.registry.notify_owner(self.parent_prop_name)

    def add_final(self, final_prog, setup_time):
        """Add a new column in the setup matrix : ie a setup that
        depends only on the final program."""
        self.__dest_default[final_prog] = setup_time

    def remove(self, initial_prog, final_prog):
        """Remove an element in the setup matrix

        Attributes:
            initial_prog -- the program at the beginig of the setup
            final_prog -- the program at the end of the setup
        """
        del self.__dest_prog[final_prog][initial_prog]
        if self.__dest_prog[final_prog]:
            del self.__dest_prog[final_prog]
        self.registry.notify_owner(self.parent_prop_name)

    def modify(self, initial_prog, final_prog, new_initial=None, new_final=None, new_time=None):
        """Change an entry in the setup matrix. If the change create a conflict..."""
        #TODO: check for duplicate keys
        time = self.__dest_prog[final_prog][initial_prog]
        if (new_initial and (not new_initial == initial_prog)) or (new_final and (not new_final == final_prog)):
            initial = new_initial or initial_prog
            final = new_final or final_prog
            logger.debug(_("changing setup: ({initial}, {final}) -> ({new_init}, {new_final})").format(initial=initial_prog,
                                                                                                       final=final_prog,
                                                                                                       new_init=initial,
                                                                                                       new_final=final))
            self.remove(initial_prog, final_prog)
            self.add(initial, final, time)
        if new_time:
            logger.debug(_("changing setup time: {0:f}) -> {1:f}").format(time, new_time))
            self.__dest_prog[final_prog][initial_prog] = new_time
        self.registry.notify_owner(self.parent_prop_name)

    def get(self, initial_prog, final_prog):
        """Get setup time. if initial or final element desn't exist in
        the matrix, the default value is returned.

        Arguments:
            initial_prog -- the program at the beginig of the setup
            final_prog -- the program at the end of the setup

        Returns:
            the setup delay
        """
        if initial_prog == final_prog:
            return 0
        if final_prog in self.__dest_prog and initial_prog in self.__dest_prog[final_prog]:
            expr = self.__dest_prog[final_prog][initial_prog]
        elif final_prog in self.__dest_default:
            expr = self.__dest_default[final_prog]
        else:
            expr = self.default_time
        return self.registry.eval_expression(expr)

    def items(self):
        """Return a list of tuple of the form (initial, final, delay)
        """
        for (final, time_elt) in self.__dest_prog.items():
            for (initial, delay) in time_elt.items():
                yield (initial, final, delay)

    def __len__(self):
        """Return the number of elements in the matrix (ie the number of setups)"""
        result = 0
        for d in self.__dest_prog.values():
            result += len(d)
        return result


class XTable(dict):
    """A dictionary of Physical Changes, where the name is the attribute to
    change and value is the new attribute value. This class is the base for
    ProgramTable and ChangeTable"""

    def __init__(self, prop_registry, parent_prop_name):
        """Create a new Xtable.

        Arguments:
            prop_registry -- the parent property Registry
            parent_prop_name -- the name of the parent property (used to notify
            of value changes)
        """
        dict.__init__(self)
        self.registry = prop_registry
        assert 'notify_owner' in dir(self.registry)
        self.parent_prop_name = parent_prop_name

    def __setitem__(self, name, value):
        """Set a value in the table"""
        dict.__setitem__(self, name, value)
        self.registry.notify_owner(self.parent_prop_name)


class ChangeTable(XTable):
    """A dictionary of Physical Changes, where the name is the attribute to
    change and value is the new attribute value"""


class ProgramTable(XTable):
    """A dictionary of Programs, designed by their names."""

    def __init__(self, prop_registry, parent_prop_name, schema):
        """Set the structure of the program's transforms
        schema is a list of tuple of the form (key, display_type, display_name)
        """
        XTable.__init__(self, prop_registry, parent_prop_name)
        self.program_keyword = schema

    def add_program(self, name, delay, prog_transform=None, prog_resources=[]):
        """Add a program in the program Table."""
        prog = Program(self.registry, delay, prog_resources)
        #initialize transforms to default values
        for (transf_name, display) in self.program_keyword:
            prog.transform[transf_name] = Display.default_value[display.type](self.registry,
                                                                              self.parent_prop_name, [])
        if prog_transform and 'items' in dir(prog_transform):
            for (transf_name, transf_value) in prog_transform.items():
                if 'items' in dir(transf_value):
                    chg_table = ChangeTable(self.registry, self.parent_prop_name)
                    prog.transform[transf_name] = chg_table
                    for (chg_name, chg_value) in transf_value.items():
                        chg_table[chg_name] = chg_value
                else:
                    prog.transform[transf_name] = transf_value
        self[name] = prog


class Program(object):
    """A program represents a transformation to apply on a product,
    and a transformation time. time can be either a numeric value, or
    a string containing a python expression that evaluate to a time.
    This is usefull to model probability distribution such as function of the
    random module. These expressions should use the globally-defined random
    number generator called rng (e.g. 'rng.expovariate(lambda=2)')

    A program is a particular form of property registry

    Attributes:
        transform -- a dictionary of program parameters
        time_law -- a python expression used to evaluate the delay (may be
                    a float, int, or an expression calling random or rng)
    """
    def __init__(self, prop_registry, time=0.0, resources=[]):
        """Create a new instance of a Program

        Arguments:
            module -- the module the own the program.
            time -- the program duration (default = 0.0). It can be either a number, or an evaluable string.
            resources -- a list of resources that the program  execution require (default [])
        """
        self.registry = prop_registry
        self.time_law = time
        self.transform = XTable(self.registry, 'program_table')
        self.resources = resources

    def time(self, product=None):
        """Return the delay corresponding to this program. If time is a string
        expression, it is evaluated to a number.
        """
        return self.registry.eval_expression(self.time_law, product)

    def is_evaluable(self):
        """Return True if time_law is evaluable"""
        class DumbProduct(object):
            """used to test whether expression evauatio would work."""
            def __getitem__(self, value):
                return 1
        product = DumbProduct()
        try:
            eval_value = self.registry.eval_expression(self.time_law, product)
        except:
            logger.exception(_("could not evaluate expression {0}").format(self.time_law))
            return False
        if not type(eval_value) == str:
            return True
        return False


class Display(object):
    """
    display_name -- the localized name that should be used to display
                            this property
            lower_bound --
            upper_bound --
    """

    REFERENCE = 1
    """A module present in the model, represented by its name"""
    VALUE = 2
    """Any (string) value"""
    BOOL_VALUE = 3
    """A boolean value"""
    INT = 4
    """A numerical value (int)"""
    FLOAT = 5
    """A numerical value (int)"""
    REFERENCE_LIST = 6
    """A list of references"""
    PROGRAM_TABLE = 7
    """A dictionary (ProgramTable) of emulica.emulation.Program, where keys are program names"""
    SETUP = 8
    """A emulica.emulation.SetupMatrix"""
    EVALUABLE = 9
    """A string that can be evaluated (using python's 'eval' function) to a numeric"""
    PHYSICAL_PROPERTIES_LIST = 10
    """a set of physical properties, and the associated value"""

    type_names = {REFERENCE: _("Module"),
                  VALUE: _("String"),
                  BOOL_VALUE: _("Boolean"),
                  INT: _("Integer"),
                  FLOAT: _("Float"),
                  REFERENCE_LIST: _("List of modules"),
                  PROGRAM_TABLE: _("Program table"),
                  SETUP: _("Setup table"),
                  EVALUABLE: _("Evaluable string"),
                  PHYSICAL_PROPERTIES_LIST: _("List of physical properties")}

    default_value = {REFERENCE: lambda reg, name, schema: None,
                     VALUE: lambda reg, name, schema: str(),
                     BOOL_VALUE: lambda reg, name, schema: bool(),
                     INT: lambda reg, name, schema: int(),
                     FLOAT: lambda reg, name, schema: float(),
                     REFERENCE_LIST: lambda reg, name, schema: list(),
                     PROGRAM_TABLE: lambda reg, name, schema: ProgramTable(reg, name, schema),
                     SETUP: lambda reg, name, schema: SetupMatrix(reg, 0, name),
                     EVALUABLE: lambda reg, name, schema: str(),
                     PHYSICAL_PROPERTIES_LIST: lambda reg, name, schema: ChangeTable(reg, name)}

    def __init__(self, prop_type, display_name=None, lower_bound=0, upper_bound=2000000):
        """Create an new instance of a ModuleProperty

        Arguments:
            prop_type -- the type of the property : either one of the int
                         constant defined in this module, or a string that is
                         the name of the constant
            value -- the value of the property
            module -- the module tha this property belong to.
        """
        self.name = display_name
        if isinstance(prop_type, int):
            self.type = prop_type
        else:
            self.type = getattr(self, prop_type)
            #self.logger.warning("")
        self.lower = lower_bound
        self.upper = upper_bound

    def is_int(self):
        """Return true if property is an integer"""
        return self.type == self.INT



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (C) 2022 Centre de Recherche en Automatique de Nancy 
#
# Authors:
# Rémi Pannequin (remi.pannequin@univ-lorraine.fr)
#
### END LICENSE


"""Try to reproduce a bug that arise when two products are added in a holder 
with presence observation.

"""

import unittest

import util
util.set_path()

import logging
from emulica.core import set_up_logging
set_up_logging(logging.WARNING)

from emulica.core.emulation import Model, Holder, PushObserver, CreateAct, Request

import logging
logger = logging.getLogger('bug1')

EMULATE_UNTIL = 100
EXP_RESULT_LEN = 1
EXP_RESULT = [] #TODO

result = []

class ObserveProduct:
    def run(self, model):
        mod = model.modules['C1_obs']
        report = mod.create_report_socket(multiple_observation=True)
        while True:
            ev = yield report.get()
            result.append(ev)


class ControlCreate:
    def run(self, model, prod_ids):
        createModule = model.modules["create"]
        report = createModule.create_report_socket(multiple_observation=True)

        for pid in prod_ids:
            m = Request("create_support",
                        "create",
                        params={'productID': pid, 'productType': 'S'})
            yield model.get_sim().timeout(5)
            yield createModule.request_socket.put(m)
            yield report.get()
            yield model.get_sim().timeout(15)


def get_model():
    model = Model()

    c1 = Holder(model, 'C1')
    c1['capacity'] = 6
    c1['speed'] = 0.5
    obs = PushObserver(model, "C1_obs", "ev_C1", holder=c1)
    obs['identify'] = True
    obs['observe_absence'] = True

    create = CreateAct(model, 'create', c1)
    model.register_control(ControlCreate, pem_args=(model, [1, 2]))
    model.register_control(ObserveProduct)
    return model


class TestBug1(unittest.TestCase):
    def setUp(self):
        print(self.id())
        
    def test_ModelCreate(self):
        get_model()

    def test_Start(self):
        model = get_model()
        model.emulate(until = EMULATE_UNTIL)

    def test_RunResults(self):
        model = get_model()
        model.emulate(until = EMULATE_UNTIL)
        self.assertEqual(len(result), EXP_RESULT_LEN)
        # self.assertEqual(result, EXP_RESULT)


if __name__ == '__main__':    
    unittest.main()





#!/usr/bin/python
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (C) 2013 Rémi Pannequin, Centre de Recherche en Automatique de Nancy remi.pannequin@univ-lorraine.fr
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

import logging
from emulica.core import set_up_logging
set_up_logging(logging.ERROR)


import unittest

import util
util.set_path()

_ = lambda x:x

from emulica.core.properties import SetupMatrix, Registry
from emulica.core import emulation



class TestSetupMatrix(unittest.TestCase):
    
    
    
    def setUp(self):
        print(self.id())
        self.model = emulation.Model()
        self.p = emulation.Product(self.model)

    def compare(self, instance, init, final, expRes):
        t = instance.get(init, final)
        self.assertEqual(t, expRes)



    def test_AddValues(self):
        instance = SetupMatrix(Registry(self.p, self.model.rng), 3)
        instance.add('p1', 'p2', 1)
        self.compare(instance, 'p1', 'p2', 1)
        instance.add('p1', 'p3', 2)
        self.compare(instance, 'p1', 'p3', 2)
        self.compare(instance, 'p0', 'p0', 0)
        self.compare(instance, 'p0', 'p1', 3)
        instance.add('p2', 'p3', 4)
        self.compare(instance, 'p2', 'p3', 4)
        self.compare(instance, 'p1', 'p2', 1)
        self.compare(instance, 'p1', 'p3', 2)
        instance.add('p3', 'p1', 5)
        self.compare(instance, 'p3', 'p1', 5)


    def test_ModValue(self):
        instance = SetupMatrix(Registry(self.p, self.model.rng), 3)
        instance.add('p3', 'p1', 5)
        instance.modify('p3', 'p1', new_final = 'p12')
        self.compare(instance, 'p3', 'p12', 5)
        instance.add('p2', 'p3', 4)
        instance.modify('p2', 'p3', new_initial = 'p12')
        self.compare(instance, 'p12', 'p3', 4)
        instance.add('p1', 'p2', 1)
        instance.modify('p1', 'p2', new_time = 2)
        self.compare(instance, 'p1', 'p2', 2)
        
        
if __name__ == '__main__':    
    unittest.main()
        
        



#!/usr/bin/python
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (C) 2013 Rémi Pannequin, Centre de Recherche en Automatique de Nancy remi.pannequin@univ-lorraine.fr
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

import unittest

import logging
from emulica.core import set_up_logging
set_up_logging(logging.ERROR)

import util
util.set_path()

from emulica.core.emulation import *
from emulica.core.properties import SetupMatrix

EXP_RESULT_PRODUCT = [(1, [(3, 7, 'machine', 'p1')], 
                          [(0, 'source'), 
                           (0, 'transporter'), 
                           (2, 'espaceMachine'), 
                           (7, 'transporter'), 
                           (10, 'sink')], 0, 80), 
                      (2, [(14, 22, 'machine', 'p3')], 
                          [(0, 'source'),
                           (10, 'transporter'), 
                           (12, 'espaceMachine'), 
                           (22, 'transporter'), 
                           (25, 'sink')], 0, 80), 
                      (3, [(28, 35, 'machine', 'p2')],
                          [(0, 'source'), 
                           (25, 'transporter'), 
                           (27, 'espaceMachine'), 
                           (35, 'transporter'), 
                           (38, 'sink')], 0, 80), 
                      (4, [(40, 45, 'machine', 'p2')], 
                          [(0, 'source'), 
                           (38, 'transporter'), 
                           (40, 'espaceMachine'), 
                           (45, 'transporter'), 
                           (48, 'sink')], 0, 80), 
                      (5, [(52, 56, 'machine', 'p1')], 
                          [(0, 'source'), 
                           (48, 'transporter'), 
                           (50, 'espaceMachine'), 
                           (56, 'transporter'), 
                           (59, 'sink')], 0, 80), 
                      (6, [(63, 71, 'machine', 'p3')], 
                          [(0, 'source'), 
                           (59, 'transporter'), 
                           (61, 'espaceMachine'), 
                           (71, 'transporter'), 
                           (74, 'sink')], 0, 80)] 
EXP_RESULT_RESOURCE = [[(0, 0, 'setup'), 
                        (0, 2, 'load'), 
                        (7, 7, 'setup'), 
                        (7, 10, 'unload'), 
                        (10, 10, 'setup'), 
                        (10, 12, 'load'), 
                        (22, 22, 'setup'), 
                        (22, 25, 'unload'), 
                        (25, 25, 'setup'), 
                        (25, 27, 'load'), 
                        (35, 35, 'setup'), 
                        (35, 38, 'unload'), 
                        (38, 38, 'setup'), 
                        (38, 40, 'load'), 
                        (45, 45, 'setup'), 
                        (45, 48, 'unload'), 
                        (48, 48, 'setup'), 
                        (48, 50, 'load'), 
                        (56, 56, 'setup'), 
                        (56, 59, 'unload'), 
                        (59, 59, 'setup'), 
                        (59, 61, 'load'), 
                        (71, 71, 'setup'), 
                        (71, 74, 'unload')], 
                       [(2, 3, 'setup'), 
                        (3, 7, 'p1'), 
                        (12, 14, 'setup'), 
                        (15, 17, 'failure'), 
                        (14, 22, 'p3'), 
                        (27, 28, 'setup'), 
                        (32, 34, 'failure'), 
                        (28, 35, 'p2'), 
                        (40, 40, 'setup'), 
                        (40, 45, 'p2'), 
                        (49, 51, 'failure'), 
                        (50, 52, 'setup'), 
                        (52, 56, 'p1'), 
                        (61, 63, 'setup'), 
                        (66, 68, 'failure'), 
                        (63, 71, 'p3')]]

EMULATE_UNTIL = 80;

class ControlCreate:
    def run(self, model):
        n = 0
        i = 0
        create = model.modules["create"]
        pType = ['type1', 'type2', 'type3', 'type3']
        while n < 6:
            m = Request("create", "create",params={'productType':pType[i]})
            yield create.request_socket.put(m)
            i = (i+1)%4
            n += 1


class ControlMachine:
    def run(self, model):
        prog = {'type1':'p1','type2':'p3','type3':'p2'}
        sp = model.modules["transporter"]
        machine = model.modules["machine"]
        rp_machine = machine.create_report_socket()
        obs1 = model.modules["obsSource"]
        rp_obs1 = obs1.create_report_socket()
        obs2 = model.modules["obsMachine"]
        rp_obs2 = obs2.create_report_socket()
        while True:
            ##attente de l'arrivée d'un pièce
            ev = yield rp_obs1.get()
            rq = Request("transporter","move",params={'program':'load'})
            yield sp.request_socket.put(rq)
            ##pièce prête
            ev = yield rp_obs2.get()
            p = prog[ev.how['productType']]
            yield machine.request_socket.put(Request("machine","setup", params={"program":p}))
            ##début process
            yield machine.request_socket.put(Request("machine","make"))
            ##attente fin process
            fin = False
            while not fin:
                e = yield rp_machine.get()
                fin = e.what=="idle"
            ##déchargement
            yield sp.request_socket.put(Request("transporter", "move", params={"program":'unload'}))


def get_model():
    model = Model()
    source = Holder(model, "source")
    sink = Holder(model, "sink")
    espaceMachine = Holder(model, "espaceMachine")

    obsSource = PushObserver(model, "obsSource", "source-ready", observe_type = False, holder = source)
    obsMachine = PushObserver(model, "obsMachine", "machine-ready", holder = espaceMachine)
    obsSink = PushObserver(model, "obsSink", "sink-ready", holder = sink)
    
    c = CreateAct(model, "create", source)
    sp = SpaceAct(model, "transporter")
    sp.add_program('load', 2, {'source':source, 'destination':espaceMachine})
    sp.add_program('unload', 3, {'source':espaceMachine, 'destination':sink})
    machine = ShapeAct(model, "machine", espaceMachine)
    machine.add_program('p1', 4)
    machine.add_program('p2', 5)
    machine.add_program('p3', 6)
    m = SetupMatrix(machine.properties, 1)
    m.add('p1','p3',2)
    m.add('p3','p1',3)
    machine['setup'] = m
    fail1 = Failure(model, "fail1", 15, 2, [machine])
    model.register_control(ControlCreate)
    model.register_control(ControlMachine)
    return model

def register_control(model):
    model.register_control(ControlCreate)
    model.register_control(ControlMachine)
    return model


class TestSim7(unittest.TestCase):
    def setUp(self):
        print(self.id())
        
    def test_ModelCreate(self):
        get_model()

    def test_Start(self):
        model = get_model()
        model.emulate(until = EMULATE_UNTIL)

    def test_RunResults(self):
        model = get_model()
        model.emulate(until = EMULATE_UNTIL)
        result_product = [(pid, p.shape_history, 
                       p.space_history,
                       p.create_time,
                       p.dispose_time) for (pid, p) in model.products.items()]
        result_resource = [model.modules["transporter"].trace, model.modules["machine"].trace]
        self.assertEqual(result_product, EXP_RESULT_PRODUCT)
        self.assertEqual(result_resource, EXP_RESULT_RESOURCE)


if __name__ == '__main__':    
    unittest.main()




#!/usr/bin/python
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (C) 2013 Rémi Pannequin, Centre de Recherche en Automatique de Nancy remi.pannequin@univ-lorraine.fr
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE
"""
This simple model based on sim1 tests the insert_request(): test of out-of-simulation request insertion
"""

import util
util.set_path()

import logging
from emulica.core import set_up_logging
set_up_logging(logging.ERROR)

import unittest



import time
import threading
from emulica.core.emulation import *
from emulica.core import controler

EXP_RESULT = [(1, [], [(5, 'holder1')], 5, 30.1),
              (2, [], [(15, 'holder1')], 15, 30.1),
              (3, [], [(30, 'holder1')], 30, 30.1)]


EMULATE_UNTIL = 30.1;


def get_model():
    model = Model()
    h = Holder(model, "holder1")
    c = CreateAct(model, "create1", h)
    o = PushObserver(model, "obs", "ev1", observe_type = False, holder = h)
    return model

class TestSim12(unittest.TestCase):
    def setUp(self):
        print(self.id())

    def test_ModelCreate(self):
        get_model()

    def test_Start(self):
        model = get_model()
        model.emulate(until = EMULATE_UNTIL)

    def begin_cb(self, model):
        #print("begin = %s"%self.begin)
        self.begin = True
    
    def finish_cb(self, model):
        self.finish = True
    
    def step(self, model):
        #print(model.sim.now)
        self.t.append(model.sim.now)
    
    def test_RunResults(self):
        model = get_model()
        finished = threading.Condition()
        self.begin = False
        self.finish = False
        self.t = []
        timer = controler.TimeControler(model, 
                                        real_time = True, 
                                        rt_factor = 1/5, 
                                        until = EMULATE_UNTIL, 
                                        step=2)
        timer.add_callback(self.begin_cb, controler.EVENT_START)
        timer.add_callback(self.finish_cb, controler.EVENT_FINISH)
        timer.add_callback(self.step, controler.EVENT_TIME)
        timer.start()
        self.assertTrue(timer.is_alive())
        time.sleep(1)
        #attente de 5s et ajout d'une requête de la simulation
        timer.dispatch(Request("create1", "create", date = 5))
        #attente 10s et ajout
        time.sleep(0.1)
        timer.dispatch(Request("create1", "create", date = 15))
        #attente 15s et ajout
        time.sleep(2)
        timer.dispatch(Request("create1", "create", date = 30))
        #attente de fin de simulation (au plus 10s)...
        timer.join(60)
        self.assertTrue(self.begin)
        self.assertTrue(self.finish)
        self.assertEqual([2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 22, 24, 26, 28, 30], self.t)
        result = [(pid, p.shape_history, p.space_history, p.create_time, p.dispose_time) for (pid, p) in model.products.items()]
        self.assertEqual(result, EXP_RESULT)

if __name__ == '__main__':
    unittest.main()




#!/usr/bin/python
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (C) 2013 Rémi Pannequin, Centre de Recherche en Automatique de Nancy remi.pannequin@univ-lorraine.fr
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

"""Submodel: model properties: linking model properties and module parameters."""

import unittest

import logging
from emulica.core import set_up_logging
set_up_logging(logging.ERROR)

import util
util.set_path()

from emulica.core.emulation import *
from emulica.core.properties import SetupMatrix, ProgramTable



#!/usr/bin/python
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (C) 2013 Rémi Pannequin, Centre de Recherche en Automatique de Nancy remi.pannequin@univ-lorraine.fr
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE


"""This model tests the random number generation features.
"""


import unittest

import logging
from emulica.core import set_up_logging
set_up_logging(logging.ERROR)

import util
util.set_path()

from emulica.core import emulation as emu
from emulica.core.emulation import Report, Request

EMULATE_UNTIL = 250;

class ControlCreate:

    def run(self, model, a):
        n = 0
        createModule = model.modules["create1"]
        while True:
            m = Request("create1", "create")
            yield createModule.request_socket.put(m)
            yield model.get_sim().timeout(model.rng.expovariate(a))
            n += 1


class ControlSpace:
    def run(self, model):
        sp = model.modules["space1"]
        obs1 = model.modules["observer1"]
        rp_obs1 = obs1.create_report_socket()
        while True:
            yield rp_obs1.get()
            rq = Request("space1","move",params={'program':'p1'})
            yield sp.request_socket.put(rq)


def get_model(a, b):
    model = emu.Model()
    h1 = emu.Holder(model, "h1")
    h2 = emu.Holder(model, "h2")
    obs1 = emu.PushObserver(model, "observer1", "ev1", holder = h1)
    obs2 = emu.PushObserver(model, "observer2", "ev2", holder = h2)
    c = emu.CreateAct(model, "create1", h1)
    sp = emu.SpaceAct(model, "space1")
    sp['setup'].default_time = 1 
    sp.add_program('p1', "rng.expovariate("+str(b)+")", {'source':h1, 'destination':h2})
    model.register_control(ControlCreate, 'run', (model, a))
    model.register_control(ControlSpace)
    return model


def run(a, b, seed, until = 250):
    model = get_model(a, b)   
    model.emulate(until, seed = seed)
    l = [(pid, p.create_time, p.dispose_time, 
          p.shape_history, p.space_history) for (pid, p) in model.products.items()]
    t = model.modules["space1"].trace
    m = model.modules["h1"].monitor.time_average()
    return (l, t, m)

class TestSim6(unittest.TestCase):
            
    def setUp(self):
        print(self.id())
        
    def test_ModelCreate(self):
        get_model(0.7, 1)

    def test_Start(self):
        run(0.7, 1, 1)

    def test_Seed1(self):
        (l1, t1, m) = run(1, 1, 123456)
        (l2, t2, m) = run(1, 1, 123456)
        self.assertEqual(t1, t2)
        self.assertEqual(l1, l2)
    
    def test_Seed2(self):
        (l1, t1, m) = run(1.2, 1, 8750)
        (l2, t2, m) = run(1.2, 1, 8750)
        self.assertEqual(t1, t2)
        self.assertEqual(l1, l2)
        (l3, t3, m) = run(1.2, 1, 480972)
        self.assertNotEqual(t1, t3)
        self.assertNotEqual(l1, l3)
    
    #remove toolong to run this test, but this may take several minutes, beware !
    def toolongtest_RunResults(self):
        e = [(0.7, 976), (0.1, 6532109)]
        s = list()
        replication = 10
        for i in range(replication):
            s.append([run(a, 1, seed+i*375, until = 20000)[2] for (a, seed) in e])

        
        result = [sum([s[i][j] for i in range(replication)])/replication for j in range(len(e))]
        exp_result = [a*a/(1-a) for (a, seed) in e]
        #print result
        #print exp_result
        for i in range(len(result)):
            self.assertAlmostEquals(result[i], exp_result[i], delta = exp_result[i]*0.05)


if __name__ == '__main__':    
    unittest.main()




#!/usr/bin/python
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (C) 2013 Rémi Pannequin, Centre de Recherche en Automatique de Nancy remi.pannequin@univ-lorraine.fr
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE


import logging
import unittest

import logging
from emulica.core import set_up_logging
set_up_logging(logging.ERROR)

import util
util.set_path()

import logging
from emulica.core import set_up_logging
set_up_logging(logging.ERROR)

#logger = logging.getLogger('emulica.emulation')
#logger.setLevel(logging.DEBUG)
# create console handler and set level to debug
#ch = logging.StreamHandler()
#ch.setLevel(logging.DEBUG)
# create formatter
#formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
# add formatter to ch
#ch.setFormatter(formatter)
# add ch to logger
#logger.addHandler(ch)


import emulica.core.emulation as emu

EXP_RESULT = [(1, [], [(0. , 'holder1')], 0., 3.),
              (2, [], [(2., 'holder1')], 2., 5.), 
              (3, [], [(4., 'holder1')], 4., 7.), 
              (4, [], [(5., 'holder1')], 5., 8.5), 
              (5, [], [(6., 'holder1')], 6., 10.), 
              (6, [], [(7., 'holder1')], 7., 11.5),
              (7, [], [(7.5, 'holder1')], 7.5, 13.),
              (8, [], [(8., 'holder1')], 8., 14.5)]
              
              
EXP_RESULT_POSITION =  [(0.5, {3.0: 1}), 
                        (1.0, {2.0: 1}), 
                        (1.5, {1.0: 1}), 
                        (2.0, {0.0: 1}), 
                        (2.5, {0.0: 1, 3.0: 2}), 
                        (3.0, {0.0: 1, 2.0: 2}), 
                        (3.5, {1.0: 2}), 
                        (4.0, {0.0: 2}), 
                        (4.5, {0: 2, 3.0: 3}), 
                        (5.0, {0: 2, 2.0: 3}), 
                        (5.5, {1.0: 3, 3.0: 4}), 
                        (6.0, {0: 3, 2.0: 4}), 
                        (6.5, {0: 3, 1: 4, 3.0: 5}), 
                        (7.0, {0: 3, 1.0: 4, 2.0: 5}), 
                        (7.5, {0: 4, 1: 5, 3.0: 6}), 
                        (8.0, {0: 4, 1: 5, 2: 6, 3: 7}), 
                        (8.5, {0: 4, 1.0: 5, 2.0: 6, 3.0: 7, 4.0:8}), 
                        (9.0, {0: 5, 1: 6, 2: 7, 3: 8}), 
                        (9.5, {0: 5, 1: 6, 2: 7, 3: 8}), 
                        (10.0, {0: 5, 1.0: 6, 2.0: 7, 3.0: 8}), 
                        (10.5, {0: 6, 1: 7, 2: 8}), 
                        (11.0, {0: 6, 1: 7, 2: 8}), 
                        (11.5, {0: 6, 1.0: 7, 2.0: 8}), 
                        (12.0, {0: 7, 1: 8}), 
                        (12.5, {0: 7, 1: 8}), 
                        (13.0, {0: 7,1.0: 8}), 
                        (13.5, {0: 8}), 
                        (14.0, {0: 8}), 
                        (14.5, {0: 8}), 
                        (15.0, {}), 
                        (15.5, {}), 
                        (16.0, {})]

pos_list = list()

EMULATE_UNTIL = 16.1;

class ControlCreate:
    def run(self, model):
        createModule = model.modules["create1"]
        requests = [emu.Request("create1", "create", date = t) for t in [0., 2., 4., 5., 6., 7., 7.5, 8.]]
        for r in requests:
            yield createModule.request_socket.put(r)
                    

class ControlDispose:
    def run(self, model):
        disposeModule = model.modules["dispose1"]
        observerModule = model.modules["observer1"]
        report = observerModule.create_report_socket()
        while True:
            yield report.get()
            yield disposeModule.request_socket.put(emu.Request("dispose1","dispose", date = model.current_time()+1))

class ObsMonitor:
    def run(self, model):
        del pos_list[:]
        obs = model.modules["observer2"]
        report = obs.create_report_socket()
        while True:
            yield obs.request_socket.put(emu.Request("observer2","observe", date = model.current_time() + 0.5))
            r = yield report.get()
            pos_list.append((r.when, r.how['ID_by_position']))
            
            
def get_model(stepping = False):
    model = emu.Model()
    h = emu.Holder(model, "holder1")
    h['capacity'] = 5
    h['speed'] = 2 #go forward 1 unit every 0.5 unit of time
    obs1 = emu.PushObserver(model, "observer1", "ev1", observe_type = False, holder = h)
    obs2 = emu.PullObserver(model, "observer2", "position", holder = h)
    c = emu.CreateAct(model, "create1", h)
    d = emu.DisposeAct(model, "dispose1", h)
    model.register_control(ControlDispose)
    model.register_control(ControlCreate)
    model.register_control(ObsMonitor)
    return model

def register_control(model):
    model.register_control(ControlDispose)
    model.register_control(ControlCreate)
    model.register_control(ObsMonitor)
    return model


class TestSim8(unittest.TestCase):
    
    def setUp(self):
        print(self.id())
    
    def test_ModelCreate(self):
        get_model()

    def test_RunResults(self):
        model = get_model()
        model.emulate(until = EMULATE_UNTIL)
        result = [(pid, 
               p.shape_history, 
               p.space_history, 
               p.create_time, 
               p.dispose_time) for (pid, p) in model.products.items()]
        self.assertEqual(result, EXP_RESULT)
        self.assertEqual(pos_list, EXP_RESULT_POSITION)


if __name__ == '__main__':    
    unittest.main()




#!/usr/bin/python
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (C) 2013 Rémi Pannequin, Centre de Recherche en Automatique de Nancy remi.pannequin@univ-lorraine.fr
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE


import util
util.set_path()

import unittest

import logging
from emulica.core import set_up_logging
set_up_logging(logging.ERROR)



import emulica.core.emulation as emu

EXP_RESULT = [(1, [], [(0, 'holder1')], 0, 4),
              (2, [], [(10, 'holder1')], 10, 14), 
              (3, [], [(20, 'holder1')], 20, 24), 
              (4, [], [(30, 'holder1')], 30, 34), 
              (5, [], [(40, 'holder1')], 40, 44), 
              (6, [], [(50, 'holder1')], 50, 54), 
              (7, [], [(60, 'holder1')], 60, 64), 
              (8, [], [(70, 'holder1')], 70, 74), 
              (9, [], [(80, 'holder1')], 80, 84), 
              (10, [], [(90, 'holder1')], 90, 94)]

EMULATE_UNTIL = 100;

#Control processes
class ControlCreate:
    def run(self, model):
        n = 0
        createModule = model.modules["create1"]
        report = createModule.create_report_socket()
        while n < 10:
            m = emu.Request("create1", "create")
            yield createModule.request_socket.put(m)
            #print "send request to create actuator"
            rp = yield report.get()
            #print(rp)
            #print "got report from create actuator"
            yield model.get_sim().timeout(10)

class ControlDispose:
    def run(self, model):
        disposeModule = model.modules["dispose1"]
        observerModule = model.modules["observer1"]
        report = observerModule.create_report_socket()
        act_report = disposeModule.create_report_socket()
        while True:
            rp = yield report.get()
            #print("got report from observer1", rp)
            #print("sending request to dispose")
            r = emu.Request(actor="dispose1",action="dispose",date=model.current_time()+4)
            #print(r)
            yield disposeModule.request_socket.put(r)
            r2 = yield act_report.get()
            #print("got report from dispose act", r2)
            
        

def get_model():
    model = emu.Model()
    h = emu.Holder(model, "holder1")
    obs1 = emu.PushObserver(model, "observer1", "ev1", observe_type = False, holder = h)
    c = emu.CreateAct(model, "create1", h)
    d = emu.DisposeAct(model, "dispose1", h)
    model.register_control(ControlCreate)
    model.register_control(ControlDispose)
    return model

def register_control(model):
    model.register_control(ControlCreate)
    model.register_control(ControlDispose)
    return model

class TestSim1(unittest.TestCase):
    """
    In this very simple example, we create the mot simple model possible:
    a create actuator put some product in a holder. These product trigger 
    a dispose actuator thanks to a product observer on the holder.
    """

    """
    def test_ModelCreate(self):
        get_model()
    
    
    def test_ModelControl(self):
        #register control processes
        model = emu.Model()
        model.register_control(ControlCreate)
        model.register_control(ControlDispose)
    """
    
    def setUp(self):
        print(self.id())
    
    def test_Start(self):
        model = get_model()
        model.emulate(until = EMULATE_UNTIL)


    def test_RunResults(self):
        model = get_model()
        
        model.emulate(until = EMULATE_UNTIL)
        result = [(pid, 
                   p.shape_history, 
                   p.space_history, 
                   p.create_time, 
                   p.dispose_time) for (pid, p) in model.products.items()]
        self.assertEqual(result, EXP_RESULT)

    """
    def test_MultipleRun(self):
        model = get_model()
        model.emulate(until = EMULATE_UNTIL)
        model.emulate(until = EMULATE_UNTIL)
        result = [(pid, 
                   p.shape_history, 
                   p.space_history, 
                   p.create_time, 
                   p.dispose_time) for (pid, p) in model.products.items()]
        self.assertEqual(result, EXP_RESULT)
    """

if __name__ == '__main__':    
    unittest.main()




#!/usr/bin/python
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (C) 2013 Rémi Pannequin, Centre de Recherche en Automatique de Nancy remi.pannequin@univ-lorraine.fr
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

"""
This simple model based on sim1 tests the insert_request() method of emulation.Model
"""

import logging
from emulica.core import set_up_logging
set_up_logging(logging.ERROR)

import unittest

import util
util.set_path()

import emulica.core.emulation as emu

EXP_RESULT = [(1, [], [(5, 'holder1')], 5, 100),
              (2, [], [(15, 'holder1')], 15, 100),
              (3, [], [(30, 'holder1')], 30, 100)]

EMULATE_UNTIL = 100;

class ControlCreate:
    def run(self, model):
        rq = emu.Request("create1", "create")
        yield model.get_sim().timeout(5)
        model.insert_request(rq)
        yield model.get_sim().timeout(10)
        model.insert_request(rq)
        yield model.get_sim().timeout(15)
        model.insert_request(rq)
        
def get_model():
    model = emu.Model()
    h = emu.Holder(model, "holder1")
    c = emu.CreateAct(model, "create1", h)
    model.register_control(ControlCreate)
    return model
    
    
class TestSim11(unittest.TestCase):
    def setUp(self):
        print(self.id())

    def test_ModelCreate(self):
        get_model()

    def test_Start(self):
        model = get_model()
        model.emulate(until = EMULATE_UNTIL)

    def test_RunResults(self):
        model = get_model()
        model.emulate(until = EMULATE_UNTIL)
        result = [(pid, 
               p.shape_history, 
               p.space_history, 
               p.create_time, 
               p.dispose_time) for (pid, p) in model.products.items()]
        self.assertEqual(result, EXP_RESULT)


if __name__ == '__main__':    
    unittest.main()




import sys
import os.path


def set_path():
#    sys.path.insert(0, os.path.realpath(os.path.join(os.path.dirname(__file__), "../emulica-core")))
    sys.path.insert(0, os.path.realpath(os.path.join(os.path.dirname(__file__), "..")))



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (C) 2013 Rémi Pannequin, Centre de Recherche en Automatique de Nancy 
# remi.pannequin@univ-lorraine.fr
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE
"""
This module enable running simulation/emulation model with various time advances
strategies (discrete-events, real-time, hybrid). It also support running emulation
model across a network.
"""


from . import emuML, emulation
import time, threading, traceback, sys
from twisted.protocols.basic import LineReceiver
from twisted.internet.protocol import Factory
from twisted.internet import reactor
import logging
import copy

EVENT_TIME = 1
EVENT_FINISH = 2
EVENT_START = 3
EXCEPTION = 4
NAME = 'emulator'
ACTIONS = ['start', 'stop', 'resume', 'pause']

logger = logging.getLogger('emulica.controler')

class TimeControler(threading.Thread):
    """This class enable to execute a simulation or emulation model in a new thread.
    Time avance mode can be either discrete-events or real-time. In real-time, 
    the simulation time advance according to the system time (multiplied 
    by a coefficient, called real-time factor) 
    
    Attributes:
        model -- the model being executed
        real_time -- boolean indicating if run mod is real_time
        rt_factor -- real time factor
        until -- time limit for simulation execution
        finished_condition -- a threading.Condition that is notified when execution stops
        finished -- boolean, True if simulation is finished
        paused -- boolean, True if model is paused
        
    Signals:
        EVENT_TIME -- signal trigerred when time change in the simulation
        EVENT_FINISH -- signal trigerred when execution finish
        EVENT_START -- signal triggered when execution begins
        EXCEPTION -- signal triggered when execution encounter an exception
                     callback handler(Exception e, traceback tb)
                     NB: the tb is the result of the extract_tb of module traceback
    """
    
    
    
    def __init__(self, model, real_time=False, rt_factor=1, until=0, step=-1):
        """Create a new instance of a TimeControler
        
        Arguments:
            model -- the model to run
            real_time -- whether or not run real-time (default = False)
            rt_factor -- real-time factor
            until -- limit time for run
            step -- if true, EVENT_TIME is called at each event. If real time is
                    true, step is forced to true. (default = False)
        """
        threading.Thread.__init__(self)
        self.model = model
        self.real_time = real_time
        self.rt_factor = rt_factor
        self.until = until
        if step == -1:
            self.step = 1 / rt_factor
        else:
            self.step = step
        self.setDaemon(True)
        self.paused = False
        self.finished = False
        self.__pause_condition = threading.Condition()
        self.__pause_delay = 0
        self.callbacks = {EVENT_TIME: [], EVENT_FINISH: [], EVENT_START: [], EXCEPTION: []}
        self.__event_condition = threading.Condition()
        
    def run(self):
        """Start the thread"""
        for c in self.callbacks[EVENT_START]:
            c(self.model)
        try:
            if self.until == 0 and not self.real_time:
                raise RunTimeError(_("a time limit must be set if executing in discrete-events mode"))#@UndefinedVariable
            self.__begin = time.time()
            self.model.emulate(self.until, 
                               rt = self.real_time, 
                               step = self.step, 
                               callback = self.__callback, 
                               rt_factor=self.rt_factor)
        except Exception as e:
            #extract line number and pass it to the handler...
            tb = traceback.extract_tb(sys.exc_info()[2])
            traceback.print_exc()
            logger.warning(e)
            logger.warning(tb)
            for handler in self.callbacks[EXCEPTION]:
                handler(e, tb)
        finally:
            for c in self.callbacks[EVENT_FINISH]:
                c(self.model)

    def __delay(self):
        """Returns system time elapsed from the begining of simulation"""
        return time.time() - self.__begin - self.__pause_delay
    
    def __correction(self):
        """Returns the real-time correction to apply"""
        delta = (self.model.sim.now / self.rt_factor) - self.__delay() - 0.002
        return max(delta, 0.0)
    
    def __callback(self):
        """Method that is called after every simulation event """
        if self.paused:
            self.__pause_condition.acquire()
            self.__pause_condition.wait()
            self.__pause_condition.release()
        #delta = self.__correction()
        #if not delta == 0:
        for c in self.callbacks[EVENT_TIME]:
            c(self.model)
        #else:
        #    logger.debug('delta == 0')
        #if self.real_time:
        #    #wait for a timeout (delta) or for an interruption
        #    self.__event_condition.acquire()
        #    self.__event_condition.wait(delta)
        #    self.__event_condition.release()
        #print "elapsed=%f, st=%f, correction=%f"%(self.__delay(), self.model.current_time(), delta)
    
    def add_callback(self, callback, event):
        """Add a callback on event
        
        Arguments:
            callback -- function to call (it will called with the model as parameter)
            event -- the event on which the callback is registered (see constants of this module)
        """
        self.callbacks[event].append(callback)
    
    def pause(self):
        """Pause emulation"""
        self.__pause_condition.acquire()
        self.__pause_begin = time.time()
        self.paused = True
        self.__pause_condition.release()
    
    def resume(self):
        """Resume emulation"""
        self.__pause_condition.acquire()
        self.__pause_delay += time.time() - self.__pause_begin
        self.paused = False
        self.__pause_condition.notify()
        self.__pause_condition.release()
    
    def stop(self):
        """Stop emulation"""
        self.model.stop()
    
    def dispatch(self, request):
        """Send a request to an emulation module."""
        if request.who == NAME:
            if request.what in ACTIONS:
                logger.debug(_("Request {0.who} to {0.what}").format(request))
                action = getattr(TimeControler, request.what)
                action(self)
            else:
                logger.warning(_("Unknown action {0}".format(request.what)))
        elif self.model.has_module(request.who):
            #insert event in model
            logger.info(_("Inserting request for module {0.who} to do {0.what} into model").format(request))
            self.model.insert_request(request)
            #interrupt sleep
            #self.__event_condition.acquire()
            #self.__event_condition.notify()
            #self.__event_condition.release()
        else:
            logger.warning(_("unknown module {0}").format(request.who))


class EmulationServer:
    """This class implements a TCP server that runs an emulation model. 
    This server can receive Requests, and sent Reports : it take requests as 
    input, and send them to the right emulation module (using a dispatch 
    method). When reports are generated by modules, it send them to clients.
    
    Attributes:
        model -- the emulation model
        factory -- twisted's Factory
    """
    
    def __init__(self, model, port, rt_factor=1.):
        """Create an instance of an EmulationServer"""
        self.factory = Factory()
        self.factory.protocol = EmulationProtocol
        self.factory.clients = dict()
        self.factory.app = self
        self.port = port
        self.model = model
        self.rt_factor = rt_factor
    
    def start(self):
        """Start the server"""
        reactor.listenTCP(self.port, self.factory)#@UndefinedVariable
        reactor.run()#@UndefinedVariable
    
    def initialize_controler(self, client):
        """Make the emulation thread ready to run"""
        #first make a deepcopy of the emulation model
        new_model = copy.deepcopy(self.model)
        #create a new controller instance and initialize it
        controler = TimeControler(new_model, real_time=True, rt_factor=self.rt_factor, until=100000, step=True)
        controler.add_callback(client.notify_stop, EVENT_FINISH)
        controler.add_callback(client.notify_start, EVENT_START)
        controler.add_callback(client.notify_time, EVENT_TIME)
        #add callback to generate emulator reports, using the ReportSource class
        controler.model.register_control(ReportSource, pem_args = (controler.model, client.send_report))
        #store client and corresponding controler
        self.factory.clients[client] = controler


class ReportSource:
    """This simPy Process is used to get Reports from emulation modules."""
    def run(self, model, send):
        """PEM : create a Store, and attach it to every module in the model"""
        r = model.new_report_socket()
        for module in model.module_list():
            module.attach_report_socket(r)
            logging.info(_("attaching reports store to module {0}").format(module.name))
        while True:
            report = yield r.get()
            send(report)


class EmulationProtocol(LineReceiver):
    """
    This class is the protocol used by the server...
    First version of the protocol: only one client can connect !
    """
    delimiter = b"\n"
    MAX_CLIENT = 100
    
    def connectionMade(self):
        """Serve a client, that has just opened a connection. 
        """
        if len(self.factory.clients) < EmulationProtocol.MAX_CLIENT:
            logging.info(_("connection opened by {0}").format(self.transport.getPeer().host))
            self.factory.app.initialize_controler(self)
            self.send_report(emulation.Report(NAME, 'ready'))
        else:
            self.sendLine(_("Another client is already connected: disconnecting..."))
            self.transport.loseConnection()
       
    def connectionLost(self, reason):
        """a connection just closed: remove protocol from clients table"""
        if self in self.factory.clients:
            del self.factory.clients[self]
    
    def lineReceived(self, line):
        """an XML message has been received"""
        try:
            request = emuML.parse_request(line)
            logging.info(_("received request {0}").format(str(request)))
            self.factory.clients[self].dispatch(request)
        except emuML.EmuMLError as message:
            logging.warning(_("Error in processing message: {0}").format(message))
    
    def notify_stop(self, model):
        """Stop the server"""
        print("Emulation finished")
        self.send_report(emulation.Report(NAME, 'finished'))
        #reactor.stop()
    
    def notify_start(self, model):
        """Notify starting of emulation"""
        self.send_report(emulation.Report(NAME, 'running'))
    
    def notify_time(self, model):
        """Notify of time advance"""
        print(model.current_time())
    
    def send_report(self, report):
        """Send a report to the client"""
        message = emuML.write_report(report)
        logging.info(_("sending report {0}").format(message))
        self.sendLine(message)
        reactor.wakeUp()#@UndefinedVariable



#!/usr/bin/python
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

### BEGIN LICENSE
# Copyright (C) 2013 Rémi Pannequin, Centre de Recherche en Automatique de Nancy remi.pannequin@univ-lorraine.fr
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE


"""
In this test model, we use assemble actuators...
*With assemble keys*
two type of products are created, and then assembled, and finally put in inventory
"""



import unittest

import util
util.set_path()

import logging
from emulica.core import set_up_logging
set_up_logging(logging.WARNING)

from emulica.core import emulation
import logging
logger = logging.getLogger('test_sim20')

EXP_RESULT_RESOURCE = [[(5, 5, 'setup'), (5, 10.0, 'p'), (13.0, 18.0, 'p'), (21.0, 26.0, 'p'), (29.0, 34.0, 'p'), (37.0, 42.0, 'p'), (45.0, 50.0, 'p'), (53.0, 58.0, 'p')], 
                       [(15.0, 15.0, 'setup'), (15.0, 20.0, 'p'), (24.0, 29.0, 'p'), (32.0, 37.0, 'p'), (40.0, 45.0, 'p'), (48.0, 53.0, 'p'), (56.0, 61.0, 'p'), (64.0, 69.0, 'p')]]


EXP_RESULT_PRODUCT = [
(1,  [( 5.0, 10.0, 'assy', 'p')], [(0, 'source1'), (0, 'trans'), (2.0, 'assy_space'), (10.0, 'trans'), (11.0, 'buffer'), (13.0, 'trans'), (15.0, 'unassy_space'), (21.0, 'trans'), (22.0, 'out1')], 0, 100.0), 
(2,  [(13.0, 18.0, 'assy', 'p')], [(1, 'source1'), (11.0, 'trans'), (13.0, 'assy_space'), (18.0, 'trans'), (19.0, 'buffer'), (22.0, 'trans'), (24.0, 'unassy_space'), (29.0, 'trans'), (30.0, 'out1')], 1, 100.0), 
(3,  [(21.0, 26.0, 'assy', 'p')], [(3, 'source1'), (19.0, 'trans'), (21.0, 'assy_space'), (26.0, 'trans'), (27.0, 'buffer'), (30.0, 'trans'), (32.0, 'unassy_space'), (37.0, 'trans'), (38.0, 'out1')], 3, 100.0), 
(4,  [( 5.0, 10.0, 'assy', 'p')], [(5, 'source2'), (5, 'assy_space'), (10.0, 'trans'), (11.0, 'buffer'), (13.0, 'trans'), (15.0, 'unassy_space'), (20.0, 'out2')], 5, 100.0), 
(5,  [(13.0, 18.0, 'assy', 'p')], [(6, 'source2'), (13.0, 'assy_space'), (18.0, 'trans'), (19.0, 'buffer'), (22.0, 'trans'), (24.0, 'unassy_space'), (29.0, 'out2')], 6, 100.0), 
(6,  [(29.0, 34.0, 'assy', 'p')], [(7, 'source1'), (27.0, 'trans'), (29.0, 'assy_space'), (34.0, 'trans'), (35.0, 'buffer'), (38.0, 'trans'), (40.0, 'unassy_space'), (45.0, 'trans'), (46.0, 'out1')], 7, 100.0), 
(7,  [(21.0, 26.0, 'assy', 'p')], [(7, 'source2'), (21.0, 'assy_space'), (26.0, 'trans'), (27.0, 'buffer'), (30.0, 'trans'), (32.0, 'unassy_space'), (37.0, 'out2')], 7, 100.0), 
(8,  [(29.0, 34.0, 'assy', 'p')], [(9, 'source2'), (29.0, 'assy_space'), (34.0, 'trans'), (35.0, 'buffer'), (38.0, 'trans'), (40.0, 'unassy_space'), (45.0, 'out2')], 9, 100.0), 
(9,  [(37.0, 42.0, 'assy', 'p')], [(11, 'source2'), (37.0, 'assy_space'), (42.0, 'trans'), (43.0, 'buffer'), (46.0, 'trans'), (48.0, 'unassy_space'), (53.0, 'out2')], 11, 100.0), 
(10, [(37.0, 42.0, 'assy', 'p')], [(12, 'source1'), (35.0, 'trans'), (37.0, 'assy_space'), (42.0, 'trans'), (43.0, 'buffer'), (46.0, 'trans'), (48.0, 'unassy_space'), (53.0, 'trans'), (54.0, 'out1')], 12, 100.0), 
(11, [(45.0, 50.0, 'assy', 'p')], [(20, 'source1'), (43.0, 'trans'), (45.0, 'assy_space'), (50.0, 'trans'), (51.0, 'buffer'), (54.0, 'trans'), (56.0, 'unassy_space'), (61.0, 'trans'), (62.0, 'out1')], 20, 100.0), 
(12, [(45.0, 50.0, 'assy', 'p')], [(23, 'source2'), (45.0, 'assy_space'), (50.0, 'trans'), (51.0, 'buffer'), (54.0, 'trans'), (56.0, 'unassy_space'), (61.0, 'out2')], 23, 100.0), 
(13, [(53.0, 58.0, 'assy', 'p')], [(30, 'source1'), (51.0, 'trans'), (53.0, 'assy_space'), (58.0, 'trans'), (59.0, 'buffer'), (62.0, 'trans'), (64.0, 'unassy_space'), (69.0, 'trans'), (70.0, 'out1')], 30, 100.0), 
(14, [(53.0, 58.0, 'assy', 'p')], [(35, 'source2'), (53.0, 'assy_space'), (58.0, 'trans'), (59.0, 'buffer'), (62.0, 'trans'), (64.0, 'unassy_space'), (69.0, 'out2')], 35, 100.0)]

EMULATE_UNTIL = 200;


class ControlCreate:
    def run(self, model):
        create1 = model.modules["create_carrier"]
        create2 = model.modules["create_part1"]
        create3 = model.modules["create_part2"]
       
        dates1 = [0, 1]
        requests1 = [
            emulation.Request("create_carrier", "create",params={'productType':'carrier'}, date=d)
            for d in dates1]
        dates2 = [0]*9
        requests2 = [
            emulation.Request("create_part1", "create",params={'productType':'P1'}, date=d)
            for d in dates2]
        requests3 = [
            emulation.Request("create_part2", "create",params={'productType':'P2'}, date=d)
            for d in dates2]
        
        for rq in requests1:
            yield create1.request_socket.put(rq)
        for rq in requests2:
            yield create2.request_socket.put(rq)
        for rq in requests3:
            yield create3.request_socket.put(rq)


class ControlAssy:
    def run(self, model):
        trans = model.modules["trans"]
        assy = model.modules["assy"]
        rp_assy = assy.create_report_socket()
        obs_in = model.modules["obs_input"]
        obs1 = model.modules["obs_source_part1"]
        obs2 = model.modules["obs_source_part2"]
        rp_obsin = obs_in.create_report_socket()
        rp_obs1 = obs1.create_report_socket()
        rp_obs2 = obs2.create_report_socket()
        obs_assy = model.modules["obs_assy"]
        rp_obs_assy = obs_assy.create_report_socket()
        n = 0
        while n < 3:
            ##attente de l'arrivée d'un pièce
            print("waiting for carrier")
            ev = yield rp_obsin.get()
            print(ev)
            print("carrier ready")
            
            print("requesting carrier admission in assy station")
            rq = emulation.Request("trans","move",params={'program':'load_assy'})
            yield trans.request_socket.put(rq)
            
            yield rp_obs_assy.get()
            print("carrier loaded in assembly station")
            ##début process
            print("assembly...")
            yield assy.request_socket.put(emulation.Request("assy","assy", params={"program":'p1_z1'}))
            yield assy.request_socket.put(emulation.Request("assy","assy", params={"program":'p2_z1'}))
            yield assy.request_socket.put(emulation.Request("assy","assy", params={"program":'p1_z1'}))
            yield assy.request_socket.put(emulation.Request("assy","assy", params={"program":'p2_z2'}))
            yield assy.request_socket.put(emulation.Request("assy","assy", params={"program":'p1_z2'}))
            yield assy.request_socket.put(emulation.Request("assy","assy", params={"program":'p2_z2'}))

            ##attente fin process
            fin = 6
            while fin > 0:
                ev = yield rp_assy.get()
                print(ev)
                if ev.what=="idle":
                    fin -= 1
            ##déchargement
            print("unloading assembly station")
            yield trans.request_socket.put(emulation.Request("trans", "move", params={"program": 'unload_assy'}))
            n += 1
    

class ControlUnassy:
    def run(self, model):
        obs = model.modules["obs_buffer"]
        rp_obs = obs.create_report_socket()
        obs_unassy = model.modules["obs_unassy"]
        rp_obs_unassy = obs_unassy.create_report_socket()
        trans = model.modules["trans"]
        unassy = model.modules["unassy"]
        rp_unassy = unassy.create_report_socket()
        n = 0
        while n < 3:
            ev = yield rp_obs.get()
            rq = emulation.Request("trans","move",params={'program':'load_unassy'})
            yield trans.request_socket.put(rq)
            # wait for product to be loaded
            yield rp_obs_unassy.get()
            ##début process
            print("unassy process")
            yield unassy.request_socket.put(emulation.Request("unassy","unassy", params={"program":'z1'}))
            yield unassy.request_socket.put(emulation.Request("unassy","unassy", params={"program":'z2'}))
            ##attente fin process
            fin = 2
            while fin > 0:
                ev = yield rp_unassy.get()
                if ev.what=="idle":
                    fin -= 1
            rq = emulation.Request("trans","move",params={'program':'unload_unassy'})
            yield trans.request_socket.put(rq)
            n += 1


def get_model():
    model = emulation.Model()
    source1 = emulation.Holder(model, "input")
    obs_source1 = emulation.PushObserver(model, "obs_input", holder = source1)
    source2 = emulation.Holder(model, "source_part1")
    obs_source2 = emulation.PushObserver(model, "obs_source_part1", holder = source2)
    source3 = emulation.Holder(model, "source_part2")
    obs_source3 = emulation.PushObserver(model, "obs_source_part2", holder = source3)

    create1 = emulation.CreateAct(model, "create_carrier", destination = source1)
    create2 = emulation.CreateAct(model, "create_part1", destination = source2)
    create3 = emulation.CreateAct(model, "create_part2", destination = source3)

    assy_space = emulation.Holder(model, "assy_space")
    unassy_space = emulation.Holder(model, "unassy_space")
    obs_assy = emulation.PushObserver(model, "obs_assy", holder = assy_space)
    assy = emulation.AssembleAct(model, "assy", assy_holder = assy_space)
    unassy = emulation.DisassembleAct(model, "unassy", unassy_holder = unassy_space)
    obs_unassy = emulation.PushObserver(model, "obs_unassy", holder = unassy_space)
    trans = emulation.SpaceAct(model, "trans")
    buff = emulation.Holder(model, "buffer")
    out1 = emulation.Holder(model, "out1")
    out2 = emulation.Holder(model, "out2")
    obs_buff = emulation.PushObserver(model, "obs_buffer", holder = buff)
    trans.add_program('load_assy', 2, {'source':source1, 'destination':assy_space})
    trans.add_program('unload_assy', 1, {'source':assy_space, 'destination':buff})
    trans.add_program('load_unassy', 2, {'source':buff, 'destination':unassy_space})
    trans.add_program('unload_unassy', 1, {'source':unassy_space, 'destination':source1})
    assy.add_program('p1_z1', 5, {'source':source2, 'key': 'z1'})
    assy.add_program('p1_z2', 5, {'source':source2, 'key': 'z2'})
    assy.add_program('p2_z1', 5, {'source':source3, 'key': 'z1'})
    assy.add_program('p2_z2', 5, {'source':source3, 'key': 'z2'})
    unassy.add_program('z1', 5, {'destination':out1, 'key': 'z1'})
    unassy.add_program('z2', 5, {'destination':out2, 'key': 'z2'})
    model.register_control(ControlCreate)
    model.register_control(ControlAssy)
    model.register_control(ControlUnassy)
    return model


class TestSim20(unittest.TestCase):
    def setUp(self):
        print(self.id())
        
    def test_ModelCreate(self):
        get_model()

    def test_Start(self):
        model = get_model()
        model.emulate(until = EMULATE_UNTIL)

    
    def test_RunResults(self):
        model = get_model()
        model.emulate(until = EMULATE_UNTIL)
        prod_out1 = [(pid, p.product_type, p.space_history[-1][1]) for (pid, p) in model.products.items()]
        
        def display_components(p):
            
            s = f"{p.product_type}(#{p.pid})"
            for k, v in p.components.items():
                if k == '':
                    s += f"->{display_components(v)}"
                else:
                    s += f" {k}: {display_components(v)}"
            return s
        out1 = []
        out2 = []
        c = []
        for p in model.modules['out1'].get_products():
            out1.append(display_components(p))
        print("in out2")
        for p in model.modules['out2'].get_products():
            out2.append(display_components(p))
        for p in [p for p in model.products.values() if p.product_type == 'carrier']:
            c.append(display_components(p))
        self.assertListEqual(out1, [
            "P1(#2)->P2(#6)->P1(#3)",
            "P1(#5)->P2(#12)->P1(#7)",
            "P1(#11)->P2(#17)->P1(#13)"])
        self.assertListEqual(out2, [
            "P2(#8)->P1(#4)->P2(#10)",
            "P2(#14)->P1(#9)->P2(#16)",
            "P2(#18)->P1(#15)->P2(#19)"])
        self.assertListEqual(c, ["carrier(#1)","carrier(#20)"])


if __name__ == '__main__':    
    unittest.main()



#!/usr/bin/python
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (C) 2013 Rémi Pannequin, Centre de Recherche en Automatique de Nancy remi.pannequin@univ-lorraine.fr
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

"""
In this model, we test if the same report socket can be created and attached to several modules.
"""

import logging
from emulica.core import set_up_logging
set_up_logging(logging.ERROR)

import unittest

import util
util.set_path()

import emulica.core.emulation as emu

EXP_REPORT_LIST = [emu.Report("create1", "create-done", date = d*10) for d in range(10)]
EXP_REPORT_LIST.insert(1, emu.Report("observer1", "ev1", location = "holder1", date = 0))
report_list = []

EMULATE_UNTIL = 100;


class ControlCreate:
    def run(self, model):
        n = 0
        createModule = model.modules["create1"]
        while n < 10:
            m = emu.Request("create1", "create")
            yield createModule.request_socket.put(m)
            yield model.get_sim().timeout(10)

class ReportMonitor:
    def run(self, model):
        """PEM : create a Store, and attach it to every module in the model"""
        queue = model.create_report_socket()
        for module in model.modules.values():
            module.attach_report_socket(queue)
        while True:
            report = yield queue.get()
            report_list.append(report)

def get_model():
    model = emu.Model()
    h = emu.Holder(model, "holder1")
    obs1 = emu.PushObserver(model, "observer1", "ev1", observe_type = False, holder = h)
    c = emu.CreateAct(model, "create1", h)
    #control processes
    model.register_control(ControlCreate)
    model.register_control(ReportMonitor)
    return model





class TestSim13(unittest.TestCase):
    
    def setUp(self):
        print(self.id())
    
    def test_ModelCreate(self):
        get_model()

    def test_Start(self):
        model = get_model()
        model.emulate(until = EMULATE_UNTIL)

    def test_RunResults(self):
        model = get_model()
        model.emulate(until = EMULATE_UNTIL)
        #print report_list
        self.assertEqual(report_list, EXP_REPORT_LIST)

if __name__ == '__main__':    
    unittest.main()




#!/usr/bin/python
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (C) 2013 Rémi Pannequin, Centre de Recherche en Automatique de Nancy remi.pannequin@univ-lorraine.fr
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE


"""In this model, the create actuator is controled so that the level of product in the queue stay constant"""


import unittest

import logging
from emulica.core import set_up_logging
set_up_logging(logging.ERROR)

#import util
#util.set_path()

from emulica.core import plot

class MockEnv:
    def __init__(self):
        self.now = 0 


class MockHolder:
    def __init__(self):
        self.env = MockEnv()
        self.monitor = plot.Monitor(self.env)
    
    def addEvent(self, t, v):
        self.env.now = t
        self.monitor.observe(v)
    
    
class TestHolderPlot(unittest.TestCase):
    
    def setUp(self):
        print(self.id())

    def test_ProcessTrace(self):
        chart = plot.HolderChart()
        h = MockHolder()
        for i in range(1, 5):
            h.addEvent(i, i*2)
        chart.t_end = 10
        t,v = chart.process_trace(h.monitor.tseries(), h.monitor.yseries())
        print(t,v)
        self.assertEqual(t, [0, 1, 1, 2, 2, 3, 3, 4, 4, 10])
        self.assertEqual(v, [0, 0, 2, 2, 4, 4, 6, 6, 8, 8])

    def test_ProcessTrace0(self):
        chart = plot.HolderChart()
        h = MockHolder()
        chart.t_end = 10
        t,v = chart.process_trace(h.monitor.tseries(), h.monitor.yseries())
        print(t,v)
        self.assertEqual(t, [0, 10])
        self.assertEqual(v, [0, 0])
        

if __name__ == '__main__':    
    unittest.main()



#!/usr/bin/python
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

### BEGIN LICENSE
# Copyright (C) 2023 Rémi Pannequin, Centre de Recherche en Automatique de Nancy remi.pannequin@univ-lorraine.fr
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE


"""
Test resource in programs
"""



import unittest

import util
util.set_path()

import logging
from emulica.core import set_up_logging
set_up_logging(logging.WARNING)

from emulica.core import emulation
import logging
logger = logging.getLogger('test_sim21')


EXP_RESULT_PRODUCT = [(1, 
                       [(2.0, 5.0, 'machine', 'p1')],
                       [(0, 'input'),
                        (0, 'trans'),
                        (2.0, 'buffer'),
                        (5.0, 'trans'), 
                        (7.0, 'sink')], 0, 200), 
                      (2, 
                       [(9.0, 12.0, 'machine', 'p1')], 
                       [(1, 'input'),
                        (7.0, 'trans'),
                        (9.0, 'buffer'),
                        (12.0, 'trans'),
                        (14.0, 'sink')], 1, 200), 
                      (3, 
                       [(16.0, 19.0, 'machine', 'p1')], 
                       [(2, 'input'), 
                        (14.0, 'trans'), 
                        (16.0, 'buffer'), 
                        (19.0, 'trans'), 
                        (21.0, 'sink')], 2, 200), 
                      (4, 
                       [(23.0, 26.0, 'machine', 'p1')], 
                       [(3, 'input'), 
                        (21.0, 'trans'), 
                        (23.0, 'buffer'), 
                        (26.0, 'trans'), 
                        (28.0, 'sink')], 3, 200), 
                      (5, 
                       [(30.0, 33.0, 'machine', 'p1')], 
                       [(4, 'input'), 
                        (28.0, 'trans'), 
                        (30.0, 'buffer'), 
                        (33.0, 'trans'), 
                        (35.0, 'sink')], 4, 200)]

EXP_RESULT_RESOURCE = [[(0, 0, 'setup'),
                        (0, 2.0, 'load'),
                        (5.0, 5.0, 'setup'), 
                        (5.0, 7.0, 'unload'), 
                        (7.0, 7.0, 'setup'), 
                        (7.0, 9.0, 'load'), 
                        (12.0, 12.0, 'setup'), 
                        (12.0, 14.0, 'unload'),
                        (14.0, 14.0, 'setup'), 
                        (14.0, 16.0, 'load'), 
                        (19.0, 19.0, 'setup'), 
                        (19.0, 21.0, 'unload'), 
                        (21.0, 21.0, 'setup'), 
                        (21.0, 23.0, 'load'),
                        (26.0, 26.0, 'setup'), 
                        (26.0, 28.0, 'unload'), 
                        (28.0, 28.0, 'setup'), 
                        (28.0, 30.0, 'load'),
                        (33.0, 33.0, 'setup'), 
                        (33.0, 35.0, 'unload')], 
                       [(2.0, 2.0, 'setup'), 
                        (2.0, 5.0, 'p1'), 
                        (9.0, 12.0, 'p1'), 
                        (16.0, 19.0, 'p1'), 
                        (23.0, 26.0, 'p1'), 
                        (30.0, 33.0, 'p1')]]



EMULATE_UNTIL = 200;


class ControlCreate:
    def run(self, model):
        create = model.modules["create"]
        dates1 = [0, 1, 2, 3, 4]
        requests = [
            emulation.Request("create_carrier", "create",params={'productType':'carrier'}, date=d)
            for d in dates1]
        for rq in requests:
            yield create.request_socket.put(rq)
        

class ControlInput:
    def run(self, model):
        obs = model.modules['obs_input'].create_report_socket()
        obs2 = model.modules['obs_buffer'].create_report_socket(multiple_observation=True)
        while True:
            # wait for a part to arrive
            yield obs.get()
            yield from emulation.execute(model, 'trans', 'move', 'load')
            # wait for the part to leave the buffer
            ev = yield obs2.get()
            while ev.how['present'] == True:
                ev = yield obs2.get()

class ControlMachine:
    def run(self, model):
        obs = model.modules['obs_buffer'].create_report_socket(multiple_observation=True)
        while True:
            ev = yield obs.get()
            if ev.how['present']:
                yield from emulation.execute(model, 'machine', 'make', 'p1')
                yield from emulation.execute(model, 'trans', 'move', 'unload')


def get_model():
    model = emulation.Model()
    source = emulation.Holder(model, "input")
    obs_source = emulation.PushObserver(model, "obs_input", holder = source)
   

    create = emulation.CreateAct(model, "create", destination = source)
   
    buff = emulation.Holder(model, "buffer")
    obs_buff = emulation.PushObserver(model, "obs_buffer", holder = buff, observe_absence=True)
    sink = emulation.Holder(model, "sink")

    trans = emulation.SpaceAct(model, "trans")
    machine = emulation.ShapeAct(model, "machine", buff)

    operator = emulation.Resource(model, 'operator')
    trans.add_program('load', 2, {'source':source, 'destination':buff}, [operator])
    trans.add_program('unload', 2, {'source':buff, 'destination':sink}, [operator])
  
    machine.add_program('p1', 3, {}, [operator])

    model.register_control(ControlCreate)
    model.register_control(ControlInput)
    model.register_control(ControlMachine)
    return model


class TestSim21(unittest.TestCase):
    def setUp(self):
        print(self.id())
        
    def test_ModelCreate(self):
        get_model()

    def test_Start(self):
        model = get_model()
        model.emulate(until = EMULATE_UNTIL)

    
    def test_RunResults(self):
        model = get_model()
        model.emulate(until = EMULATE_UNTIL)
        result_product = [(pid, 
                       p.shape_history, 
                       p.space_history,
                       p.create_time, 
                       p.dispose_time) for (pid, p) in model.products.items()]
        result_resource = [model.get_module("trans").trace, model.get_module("machine").trace]
        self.assertEqual(result_product, EXP_RESULT_PRODUCT)
        self.assertEqual(result_resource, EXP_RESULT_RESOURCE)

        


if __name__ == '__main__':    
    unittest.main()
